"""
Unit tests for the errors.astrometric module.
"""

import numpy as np

from pygaia.errors import astrometric as astrom


class TestErrorsAstrometric(np.testing.TestCase):
    def test_parallax_uncertainty(self):
        """
        Verify that the function works for reasonable combinations of array and scalar
        input parameters.
        """
        self.assertTrue(astrom.parallax_uncertainty(15.0) > 0.0)
        self.assertTrue(astrom.parallax_uncertainty(15.0, release="dr4") > 0.0)
        self.assertTrue(astrom.parallax_uncertainty(15.0, release="dr3") > 0.0)
        self.assertTrue(astrom.parallax_uncertainty(15.0, release="dr5") > 0.0)
        gmags = np.linspace(6, 20, 100)
        errorsdr4 = astrom.parallax_uncertainty(gmags)
        errorsdr3 = astrom.parallax_uncertainty(gmags, release="dr3")
        errorsdr5 = astrom.parallax_uncertainty(gmags, release="dr5")
        for errordr4, errordr3, errordr5 in zip(errorsdr4, errorsdr3, errorsdr5):
            self.assertTrue(errordr4 > 0.0)
            self.assertTrue(errordr3 > errordr4)
            self.assertTrue(errordr5 < errordr4)

    def test_position_uncertainty(self):
        """
        Verify that the function works for reasonable combinations of array and scalar
        input parameters.
        """
        sigalphastar, sigdelta = astrom.position_uncertainty(15.0)
        self.assertTrue(sigalphastar > 0.0)
        self.assertTrue(sigdelta > 0.0)
        sigalphastar, sigdelta = astrom.position_uncertainty(15.0, release="dr3")
        self.assertTrue(sigalphastar > 0.0)
        self.assertTrue(sigdelta > 0.0)
        sigalphastar, sigdelta = astrom.position_uncertainty(15.0, release="dr5")
        self.assertTrue(sigalphastar > 0.0)
        self.assertTrue(sigdelta > 0.0)

        gmags = np.linspace(6, 20, 100)
        sigalphastardr4, sigdeltadr4 = astrom.position_uncertainty(gmags)
        sigalphastardr3, sigdeltadr3 = astrom.position_uncertainty(gmags, release="dr3")
        sigalphastardr5, sigdeltadr5 = astrom.position_uncertainty(gmags, release="dr5")
        for errordr4, errordr3, errordr5 in zip(
            sigalphastardr4, sigalphastardr3, sigalphastardr5
        ):
            self.assertTrue(errordr4 > 0.0)
            self.assertTrue(errordr3 > errordr4)
            self.assertTrue(errordr5 < errordr4)
        for errordr4, errordr3, errordr5 in zip(sigdeltadr4, sigdeltadr3, sigdeltadr5):
            self.assertTrue(errordr4 > 0.0)
            self.assertTrue(errordr3 > errordr4)
            self.assertTrue(errordr5 < errordr4)

    def test_proper_motion_uncertainty(self):
        """
        Verify that the function works for reasonable combinations of array and scalar
        input parameters.
        """
        sigpmra, sigpmdec = astrom.proper_motion_uncertainty(15.0)
        self.assertTrue(sigpmra > 0.0)
        self.assertTrue(sigpmdec > 0.0)
        sigpmra, sigpmdec = astrom.proper_motion_uncertainty(15.0, release="dr3")
        self.assertTrue(sigpmra > 0.0)
        self.assertTrue(sigpmdec > 0.0)
        sigpmra, sigpmdec = astrom.proper_motion_uncertainty(15.0, release="dr5")
        self.assertTrue(sigpmra > 0.0)
        self.assertTrue(sigpmdec > 0.0)

        gmags = np.linspace(6, 20, 100)
        sigpmradr4, sigpmdecdr4 = astrom.proper_motion_uncertainty(gmags)
        sigpmradr3, sigpmdecdr3 = astrom.proper_motion_uncertainty(gmags, release="dr3")
        sigpmradr5, sigpmdecdr5 = astrom.proper_motion_uncertainty(gmags, release="dr5")
        for errordr4, errordr3, errordr5 in zip(sigpmradr4, sigpmradr3, sigpmradr5):
            self.assertTrue(errordr4 > 0.0)
            self.assertTrue(errordr3 > errordr4)
            self.assertTrue(errordr5 < errordr4)
        for errordr4, errordr3, errordr5 in zip(sigpmdecdr4, sigpmdecdr3, sigpmdecdr5):
            self.assertTrue(errordr4 > 0.0)
            self.assertTrue(errordr3 > errordr4)
            self.assertTrue(errordr5 < errordr4)



"""
Provides functions for simulation the astrometric uncertainties in the Gaia catalogue
data for Gaia DR3+. The code reproduces the uncertainty model described on the `Gaia science performance pages <https://www.cosmos.esa.int/web/gaia/science-performance#astrometric%20performance>`_.
"""
import numpy as np

from pygaia.errors.utils import calc_z_plx

__all__ = [
    "parallax_uncertainty",
    "position_uncertainty",
    "proper_motion_uncertainty",
    "total_position_uncertainty",
    "total_proper_motion_uncertainty",
]

# Scaling factors for sky averaged position and proper motion uncertainties. The
# uncertainties are scaled with respect to the parallax uncertainty values. Note that
# the uncertainties are quoted in true arc terms (using phi*) for the longitude-like
# component.
_scaling_for_positions = {
    "dr3": {"Total": 0.75, "AlphaStar": 0.80, "Delta": 0.70},
    "dr4": {"Total": 0.75, "AlphaStar": 0.80, "Delta": 0.70},
    "dr5": {"Total": 0.75, "AlphaStar": 0.80, "Delta": 0.70},
}
_scaling_for_proper_motions = {
    "dr3": {"Total": 0.96, "AlphaStar": 1.03, "Delta": 0.89},
    "dr4": {"Total": 0.54, "AlphaStar": 0.58, "Delta": 0.50},
    "dr5": {"Total": 0.27, "AlphaStar": 0.29, "Delta": 0.25},
}

# Scaling factors for observation time interval with respect to (E)DR3: sqrt(33.12/59)
# for DR4, sqrt(33.12/119) for DR5, for DR4 and DR5 mission lengths of 60 and 120 months
# (accounting for the first month not being used in the astrometric solution).  Proper
# motion precisions scale as t**(-1.5). The extra factor 1/t is included in the scaling
# factors above.
#
# The assumed mission lengths are compatible with the assumptions on the science
# performance pages.
#
# The predictions for DR4 and DR5 are based on the (E)DR3 uncertainties, where the
# latter are inflated by a 'science margin' of 10 percent (factor 1.1).
_t_factor = {"dr3": 1.0, "dr4": 0.749, "dr5": 0.527}
_supported_releases = list(_t_factor.keys())
_default_release = "dr4"


def _check_release(release):
    """
    Check if the release requested by the user is supported.

    Parameters
    ----------
    release : str
        Release for which performance predictions are requested.

    Raises
    ------
    ValueError
        When an invalid string is specified for the release parameter.
    """
    if not (release in _supported_releases):
        raise ValueError("Release must be one of dr3, dr4, dr5, dr2")


def parallax_uncertainty(gmag, release=_default_release):
    """
    Calculate the sky averaged parallax uncertainty as a function of G, for a given Gaia data release.

    Parameters
    ----------
    gmag : float or float array
        Value(s) of G-band magnitude.
    release : str
        Specify the Gaia data release for which the performance is to be simulated.
        'dr3' -> Gaia (E)DR3, 'dr4' -> Gaia DR4, 'dr5' -> Gaia DR5. Default is 'dr4'.

    Returns
    -------
    parallax_uncertainty: float or array
        The parallax uncertainty in micro-arcseconds.

    Raises
    ------
    ValueError
        When an invalid string is specified for the release parameter.
    """
    _check_release(release)

    z = calc_z_plx(gmag)
    return np.sqrt(40 + 800 * z + 30 * z * z) * _t_factor[release]


def position_uncertainty(gmag, release=_default_release):
    """
    Calculate the sky averaged position uncertainties from G, for a given Gaia data release.

    Parameters
    ----------
    gmag : float or array
        Value(s) of G-band magnitude.
    release : str
        Specify the Gaia data release for which the performance is to be simulated.
        'dr3' -> Gaia (E)DR3, 'dr4' -> Gaia DR4, 'dr5' -> Gaia DR5. Default is 'dr4'.

    Returns
    -------
    ra_cosdelta_uncertainty, delta_uncertainty : float or array
        The uncertainty in alpha* and the uncertainty in delta, in that order, in
        micro-arcseconds.

    Raises
    ------
    ValueError
        When an invalid string is specified for the release parameter.

    Notes
    -----
    The uncertainties are for sky positions in the ICRS (i.e., right ascension,
    declination). Make sure your simulated astrometry is also on the ICRS.
    """
    _check_release(release)
    plx_unc = parallax_uncertainty(gmag, release=release)
    return (
        _scaling_for_positions[release]["AlphaStar"] * plx_unc,
        _scaling_for_positions[release]["Delta"] * plx_unc,
    )


def proper_motion_uncertainty(gmag, release=_default_release):
    """
    Calculate the sky averaged proper motion uncertainties from G, for a given Gaia data
    release.

    Parameters
    ----------
    gmag : float or array
        Value(s) of G-band magnitude.
    release : str
        Specify the Gaia data release for which the performance is to be simulated.
        'dr3' -> Gaia (E)DR3, 'dr4' -> Gaia DR4, 'dr5' -> Gaia DR5. Default is 'dr4'.

    Returns
    -------
    mualpha_cosdelta_uncertainty, mudelta_uncertainty : float or array
        The uncertainty in mu_alpha* and the uncertainty in mu_delta, in that order, in
        micro-arcseconds/year.

    Raises
    ------
    ValueError
        When an invalid string is specified for the release parameter.

    Notes
    -----
    The uncertainties are for proper motions in the ICRS (i.e., right ascension,
    declination). Make sure your simulated astrometry is also on the ICRS.
    """
    _check_release(release)
    plx_unc = parallax_uncertainty(gmag, release=release)
    return (
        _scaling_for_proper_motions[release]["AlphaStar"] * plx_unc,
        _scaling_for_proper_motions[release]["Delta"] * plx_unc,
    )


def total_position_uncertainty(gmag, release=_default_release):
    """
    Calculate the sky averaged total position uncertainty as a function of G and for the
    given Gaia data release.  This refers to the semi-major axis of the position error
    ellipse.

    Parameters
    ----------
    gmag : float or array
        Value(s) of G-band magnitude.

    release : str
        Specify the Gaia data release for which the performance is to be simulated.
        'dr3' -> Gaia (E)DR3, 'dr4' -> Gaia DR4, 'dr5' -> Gaia DR5. Default is 'dr4'.

    Returns
    -------
    position_uncertainty
        The semi-major axis of the position error ellipse in micro-arcseconds.

    Raises
    ------
    ValueError
        When an invalid string is specified for the release parameter.

    Notes
    -----
    The uncertainties are for positions in the ICRS (i.e., right ascension,
    declination). Make sure your simulated astrometry is also on the ICRS.
    """
    _check_release(release)
    plx_unc = parallax_uncertainty(gmag, release=release)
    return _scaling_for_positions[release]["Total"] * plx_unc


def total_proper_motion_uncertainty(gmag, release=_default_release):
    """
    Calculate the sky averaged total proper motion uncertainty as a function of G and
    for the given Gaia data release. This refers to the semi-major axis of the proper
    motion error ellipse.

    Parameters
    ----------
    gmag : float or array
        Value(s) of G-band magnitude.
    release : str
        Specify the Gaia data release for which the performance is to be simulated.
        'dr3' -> Gaia (E)DR3, 'dr4' -> Gaia DR4, 'dr5' -> Gaia DR5. Default is 'dr4'.

    Returns
    -------
    proper_motion_uncertainty
        The semi-major axis of the proper motion error ellipse in micro-arcseconds/year.

    Raises
    ------
    ValueError
        When an invalid string is specified for the release parameter.

    Notes
    -----
    The uncertainties are for proper motions in the ICRS (i.e., right ascension,
    declination). Make sure your simulated astrometry is also on the ICRS.
    """
    _check_release(release)
    plx_unc = parallax_uncertainty(gmag, release=release)
    return _scaling_for_proper_motions[release]["Total"] * plx_unc



"""
Provides tools for coordinate transformation and epoch propagation.
"""
import numpy as np
from enum import Enum, auto

from pygaia.astrometry.constants import au_km_year_per_sec
from pygaia.astrometry.vectorastrometry import (
    spherical_to_cartesian,
    cartesian_to_spherical,
    elementary_rotation_matrix,
    normal_triad,
)

__all__ = [
    "CoordinateTransformation",
    "Transformations",
    "EpochPropagation",
    "angular_distance",
]

# Obliquity of the Ecliptic (arcsec expressed in radians)
_obliquityOfEcliptic = np.deg2rad(84381.41100 / 3600.0)

# Galactic pole in ICRS coordinates (see Hipparcos Explanatory Vol 1 section 1.5, and
# Murray, 1983, # section 10.2)
_alphaGalPole = np.deg2rad(192.85948)
_deltaGalPole = np.deg2rad(27.12825)

# The galactic longitude of the ascending node of the galactic plane on the equator of
# ICRS (see Hipparcos Explanatory Vol 1 section 1.5, and Murray, 1983, section 10.2)
_omega = np.deg2rad(32.93192)

# Rotation matrix for the transformation from ICRS to Galactic coordinates. See equation
# (4.25) in chapter 4.5 of "Astrometry for Astrophysics", 2012, van Altena et al.
_matA = elementary_rotation_matrix("z", np.pi / 2.0 + _alphaGalPole)
_matB = elementary_rotation_matrix("x", np.pi / 2.0 - _deltaGalPole)
_matC = elementary_rotation_matrix("z", -_omega)
_rotationMatrixIcrsToGalactic = np.dot(_matC, np.dot(_matB, _matA))

# Alternative way to calculate the rotation matrix from ICRS to Galactic coordinates.
# First calculate the vectors describing the Galactic coordinate reference frame
# expressed within the ICRS.
#
# _vecN = array([0,0,1])
# _vecG3 = array([np.cos(_alphaGalPole)*np.cos(_deltaGalPole),
#       np.sin(_alphaGalPole)*np.cos(_deltaGalPole), np.sin(_deltaGalPole)])
# _vecG0 = np.cross(_vecN,_vecG3)
# _vecG0 = _vecG0/np.sqrt(np.dot(_vecG0,_vecG0))
# _vecG1 = -np.sin(_omega)*np.cross(_vecG3,_vecG0)+np.cos(_omega)*_vecG0
# _vecG2 = np.cross(_vecG3,_vecG1)
# _rotationMatrixIcrsToGalactic=array([_vecG1,_vecG2,_vecG3])

# Rotation matrix for the transformation from Galactic to ICRS coordinates.
_rotationMatrixGalacticToIcrs = np.transpose(_rotationMatrixIcrsToGalactic)

# Rotation matrix for the transformation from Ecliptic to ICRS coordinates.
_rotationMatrixEclipticToIcrs = elementary_rotation_matrix(
    "x", -1 * _obliquityOfEcliptic
)

# Rotation matrix for the transformation from ICRS to Ecliptic coordinates.
_rotationMatrixIcrsToEcliptic = np.transpose(_rotationMatrixEclipticToIcrs)

# Rotation matrix for the transformation from Galactic to Ecliptic coordinates.
_rotationMatrixGalacticToEcliptic = np.dot(
    _rotationMatrixIcrsToEcliptic, _rotationMatrixGalacticToIcrs
)

# Rotation matrix for the transformation from Ecliptic to Galactic coordinates.
_rotationMatrixEclipticToGalactic = np.transpose(_rotationMatrixGalacticToEcliptic)


class Transformations(Enum):
    """
    Enumeration with the available coordinate tranformations.
    """

    GAL2ICRS = ("Galactic", "ICRS")
    """
    Transform from Galactic to ICRS coordinates.
    """
    ICRS2GAL = ("ICRS", "Galactic")
    """
    Transform from ICRS to Galactic coordinates.
    """
    ECL2ICRS = ("Ecliptic", "ICRS")
    """
    Transform from Ecliptic to ICRS coordinates.
    """
    ICRS2ECL = ("ICRS", "Ecliptic")
    """
    Transform from ICRS to Ecliptic coordinates.
    """
    GAL2ECL = ("Galactic", "Ecliptic")
    """
    Transform from Galactic to Ecliptic coordinates.
    """
    ECL2GAL = ("Ecliptic", "Galactic")
    """
    Transform from Ecliptic to Galactic coordinates.
    """

    def __init__(self, fromsys, tosys):
        self.fromsystem = fromsys
        self.tosystem = tosys

    @property
    def from_system(self):
        """
        Coordinate system from which the transformation starts.

        Returns
        -------
        fromsystem : str
            String representation of the starting system.
        """
        return self.fromsystem

    @property
    def to_system(self):
        """
        Target coordinate system of the transformation.

        Returns
        -------
        tosystem : str
            String representation of the target system.
        """
        return self.tosystem


_rotation_matrix_dict = {
    Transformations.GAL2ICRS: _rotationMatrixGalacticToIcrs,
    Transformations.ICRS2GAL: _rotationMatrixIcrsToGalactic,
    Transformations.ECL2ICRS: _rotationMatrixEclipticToIcrs,
    Transformations.ICRS2ECL: _rotationMatrixIcrsToEcliptic,
    Transformations.GAL2ECL: _rotationMatrixGalacticToEcliptic,
    Transformations.ECL2GAL: _rotationMatrixEclipticToGalactic,
}


def angular_distance(phi1, theta1, phi2, theta2, return_posangle=False):
    """
    Calculate the angular distance between pairs of sky coordinates.

    Parameters
    ----------
    phi1 : float
        Longitude of first coordinate (radians).
    theta1 : float
        Latitude of first coordinate (radians).
    phi2 : float
        Longitude of second coordinate (radians).
    theta2 : float
        Latitude of second coordinate (radians).
    return_posangle : boolean
        If true return the position angle of the point (phi2, theta2), as defined in the
        tangent plane at (phi1, theta1). Where the position angle is measured from local
        north over east and ranges between 0 and 2*pi. If the distance between the two
        points is 0 or pi to within floating point precision (as given by
        np.finfo(float).resolution) then 0 is returned for the position angle. This is a
        choice for the cases where mathematically the position angle is not defined.
        This resolution (1e-15) corresponds to about 0.2 nanoarcseconds.

    Returns
    -------
    angdist : float
        Angular distance in radians.
    posangle : float
        Position angle as defined above in radians.
    """
    # The Formula below is numerically more stable than np.arccos( np.sin(theta1)*np.sin(theta2) + np.cos(phi2-phi1)*np.cos(theta1)*np.cos(theta2) ). See: https://en.wikipedia.org/wiki/Great-circle_distance
    dist = np.arctan2(
        np.sqrt(
            (np.cos(theta2) * np.sin(phi2 - phi1)) ** 2
            + (
                np.cos(theta1) * np.sin(theta2)
                - np.sin(theta1) * np.cos(theta2) * np.cos(phi2 - phi1)
            )
            ** 2
        ),
        np.sin(theta1) * np.sin(theta2)
        + np.cos(phi2 - phi1) * np.cos(theta1) * np.cos(theta2),
    )
    if not return_posangle:
        return dist
    else:
        posangle = np.arctan2(
            np.cos(theta2) * np.sin(phi2 - phi1),
            np.cos(theta1) * np.sin(theta2)
            - np.sin(theta1) * np.cos(theta2) * np.cos(phi2 - phi1),
        )
        posangle = np.where(
            np.isclose(np.sin(dist), 0, atol=np.finfo(float).resolution), 0.0, posangle
        )
        return dist, np.squeeze(np.where(posangle < 0, posangle + 2 * np.pi, posangle))


class CoordinateTransformation:
    """
    Provides methods for carrying out transformations between different coordinate
    (reference) systems.  Transformations between the following systems are supported:

    * ICRS or Equatorial (right ascension, declination)
    * Galactic (longitude, latitude)
    * Ecliptic (longitude, latitude)

    The transformations can be applied to sky coordinates or Cartesian coordinates.

    .. note::
        The transformations here are only intended for systems that represent a rotated
        version of the ICRS. For transformations from and to non-ICRS systems (such as
        FK5) refer to Astropy's `coordinates
        <https://docs.astropy.org/en/stable/coordinates/index.html>`_ package.

    Attributes
    ----------
    transformation : Transformations
        The transformation represented by the instance of this class.
    rotationMatrix : array
        Rotation matrix corresponding to the transformation.
    """

    def __init__(self, desired_transformation):
        """
        Class constructor/initializer. Sets the type of transformation to be used.

        Parameters
        ----------
        desiredTransformation : Transformations
            The kind of coordinate transformation that should be provided. For example
            Transformations.GAL2ECL.
        """
        self.transformation = desired_transformation
        self.rotationMatrix = _rotation_matrix_dict[desired_transformation]

    def start_coordinates(self):
        """
        Query which is the starting coordinate system of this transformation.

        Returns
        -------
        from_sys : str
            String with starting coordinate system.
        """
        return self.transformation.from_system

    def target_coordinates(self):
        """
        Query which is the target coordinate system of this transformation.

        Returns
        -------
        to_sys : str
            String with target coordinate system.
        """
        return self.transformation.to_system

    def transform_cartesian_coordinates(self, x, y, z):
        """
        Rotate Cartesian coordinates from one reference system to another.

        Use the rotation matrix with which the class was initialized. The inputs can be
        scalars or 1-dimensional numpy arrays. The length scale units of the inputs can
        be arbitrary but should of course be consistent.

        Parameters
        ----------
        x : float or array
            Value of X-coordinate in original reference system.
        y : float or array
            Value of Y-coordinate in original reference system.
        z : float or array
            Value of Z-coordinate in original reference system.

        Returns
        -------
        xrot : float or array
            Value of X-coordinate after rotation
        yrot : float or array
            Value of Y-coordinate after rotation
        zrot : float or array
            Value of Z-coordinate after rotation
        """
        xrot, yrot, zrot = np.dot(self.rotationMatrix, [x, y, z])
        return xrot, yrot, zrot

    def transform_sky_coordinates(self, phi, theta):
        """
        Convert sky coordinates from one reference system to another.

        Use of the rotation matrix with which the class was initialized. Inputs can be
        scalars or 1-dimensional numpy arrays.

        Parameters
        ----------
        phi : float or array
            Value of the longitude-like angle (right ascension, longitude) in radians.
        theta : float or array
            Value of the latitude-like angle (declination, latitude) in radians.

        Returns
        -------
        phirot : float or array
            Value of the transformed longitude-like angle in radians.
        thetarot : float or array
            Value of the transformed latitude-like angle in radians.
        """
        r = np.ones_like(phi)
        x, y, z = spherical_to_cartesian(r, phi, theta)
        xrot, yrot, zrot = self.transform_cartesian_coordinates(x, y, z)
        r, phirot, thetarot = cartesian_to_spherical(xrot, yrot, zrot)
        return phirot, thetarot

    def transform_proper_motions(self, phi, theta, muphistar, mutheta):
        """
        Converts proper motions from one reference system to another.

        Use the prescriptions in section 1.5.3 of the `Hipparcos Explanatory Volume 1
        <https://www.cosmos.esa.int/documents/532822/552851/vol1_all.pdf/99adf6e3-6893-4824-8fc2-8d3c9cbba2b5>`_
        (equations 1.5.18, 1.5.19).

        Parameters
        ----------
        phi : float or array
            The longitude-like angle of the position of the source (radians).
        theta : float or array
            The latitude-like angle of the position of the source (radians).
        muphistar : float or array
            Value of the proper motion in the longitude-like angle, multiplied by
            cos(latitude).
        mutheta : float or array
            Value of the proper motion in the latitude-like angle.

        Returns
        -------
        muphistarrot : float or array
            Value of the transformed proper motion in the longitude-like angle
            (including the cos(latitude) factor).
        muthetarot : float or array
            Value of the transformed proper motion in the latitude-like angle.
        """
        c, s = self._get_jacobian(phi, theta)
        return c * muphistar + s * mutheta, c * mutheta - s * muphistar

    def transform_sky_coordinate_errors(
        self, phi, theta, sigphistar, sigtheta, rho_phi_theta=0
    ):
        """
        Converts the sky coordinate uncertainties from one reference system to another,
        including the covariance term.

        Equations (1.5.4) and (1.5.20) from section 1.5 in the `Hipparcos Explanatory
        Volume 1
        <https://www.cosmos.esa.int/documents/532822/552851/vol1_all.pdf/99adf6e3-6893-4824-8fc2-8d3c9cbba2b5>`_
        are used.

        Parameters
        ----------
        phi : float or array
            The longitude-like angle of the position of the source (radians).
        theta : float or array
            The latitude-like angle of the position of the source (radians).
        sigphistar : float or array
            Uncertainty in the longitude-like angle of the position of the source
            (radians or sexagesimal units, including cos(latitude) term)
        sigtheta : float or array
            Uncertainty in the latitude-like angle of the position of the source
            (radians or sexagesimal units)
        rho_phi_theta : float or array
            Correlation coefficient of the position uncertainties. Zero by default

        Returns
        -------
        sigPhiRotStar : float or array
            The transformed uncertainty in the longitude-like angle (including
            cos(latitude) factor)
        sigThetaRot : float or array
            The transformed uncertainty in the latitude-like angle.
        rhoPhiThetaRot : float or array
            The transformed correlation coefficient.
        """
        if np.isscalar(rho_phi_theta) and not np.isscalar(sigtheta):
            rho_phi_theta = np.zeros_like(sigtheta) + rho_phi_theta
        c, s = self._get_jacobian(phi, theta)
        csqr = c * c
        ssqr = s * s
        covar = sigphistar * sigtheta * rho_phi_theta
        varphistar = sigphistar * sigphistar
        vartheta = sigtheta * sigtheta
        varphistar_rot = csqr * varphistar + ssqr * vartheta + 2.0 * covar * c * s
        vartheta_rot = ssqr * varphistar + csqr * vartheta - 2.0 * covar * c * s
        covar_rot = (csqr - ssqr) * covar + c * s * (vartheta - varphistar)
        return (
            np.sqrt(varphistar_rot),
            np.sqrt(vartheta_rot),
            covar_rot / np.sqrt(varphistar_rot * vartheta_rot),
        )

    def transform_proper_motion_errors(
        self, phi, theta, sigmuphistar, sigmutheta, rho_muphi_mutheta=0
    ):
        """
        Converts the proper motion uncertainties from one reference system to another,
        including the covariance term.

        Equations (1.5.4) and (1.5.20) from section 1.5 in the `Hipparcos Explanatory
        Volume 1
        <https://www.cosmos.esa.int/documents/532822/552851/vol1_all.pdf/99adf6e3-6893-4824-8fc2-8d3c9cbba2b5>`_
        are used.

        Parameters
        ----------
        phi : float or array
            The longitude-like angle of the position of the source (radians).
        theta : float or array
            The latitude-like angle of the position of the source (radians).
        sigmuphistar : float or attay
            Uncertainty in the proper motion in the longitude-like direction
            (including cos(latitude) factor).
        sigmutheta : float or array
            Uncertainty in the proper motion in the latitude-like direction.
        rho_muphi_mutheta : float or array
            Correlation coefficient of the proper motion errors. Zero by default.

        Returns
        -------
        sigMuPhiRotStar : float or array
            The transformed uncertainty in the proper motion in the longitude direction
            (including np.cos(latitude) factor).
        sigMuThetaRot : float or array
            The transformed uncertainty in the proper motion in the longitude direction.
        rhoMuPhiMuThetaRot : float or array
            The transformed correlation coefficient.
        """
        return self.transform_sky_coordinate_errors(
            phi, theta, sigmuphistar, sigmutheta, rho_phi_theta=rho_muphi_mutheta
        )

    def transform_covariance_matrix(self, phi, theta, covmat):
        """
        Transform the astrometric covariance matrix to its representation in the new
        coordinate system.

        Parameters
        ----------
        phi : float
            The longitude-like angle of the position of the source (radians).
        theta : float
            The latitude-like angle of the position of the source (radians).
        covmat : array
            Covariance matrix (5,5) of the astrometric parameters.

        Returns
        -------
        covmat_rot : array
            Covariance matrix in its representation in the new coordinate system.
        """

        c, s = self._get_jacobian(phi, theta)
        jacobian = np.identity(5)
        jacobian[0][0] = c
        jacobian[1][1] = c
        jacobian[3][3] = c
        jacobian[4][4] = c
        jacobian[0][1] = s
        jacobian[1][0] = -s
        jacobian[3][4] = s
        jacobian[4][3] = -s

        return np.dot(np.dot(jacobian, covmat), jacobian.T)

    def _get_jacobian(self, phi, theta):
        r"""
        Calculates the Jacobian for the transformation of the position and proper motion
        uncertainties between coordinate systems.

        This Jacobian is also the rotation matrix for the transformation of proper
        motions. See section 1.5.3 of the `Hipparcos Explanatory Volume 1
        <https://www.cosmos.esa.int/documents/532822/552851/vol1_all.pdf/99adf6e3-6893-4824-8fc2-8d3c9cbba2b5>`_
        (equation 1.5.20). This matrix has the following form:

        .. math::

            J = \begin{pmatrix}
                c & s \\
                -s & c \\
            \end{pmatrix}

        Parameters
        ----------
        phi : float or array
            The longitude-like angle of the position of the source (radians).
        theta : float or array
            The latitude-like angle of the position of the source (radians).

        Returns
        -------
        c, s : float or array
            The Jacobian matrix elements c and s corresponding to (phi, theta) and the
            currently desired coordinate system transformation.
        """

        p, q, r = normal_triad(phi, theta)

        # z_rot = z-axis of new coordinate system expressed in terms of old system
        z_rot = self.rotationMatrix[2, :]

        if p.ndim == 2:
            N = p.shape[1]
            z_rot_all = np.tile(z_rot, N).reshape(N, 3)
            p_rot = np.cross(z_rot_all, r.T)
            p_rot = p_rot / np.linalg.norm(p_rot, axis=1)[:, None]
            c = np.sum(p_rot * p.T, axis=1)
            s = np.sum(p_rot * q.T, axis=1)

            return c, s
        else:
            p_rot = np.cross(z_rot, r.T)
            p_rot = p_rot / np.linalg.norm(p_rot)
            return np.dot(p_rot, p), np.dot(p_rot, q)


class EpochPropagation:
    """
    Provides methods for transforming the astrometry and radial velocity of a given
    source to a different epoch.

    The formulae for rigorous epoch transformation can be found on the `Gaia
    documentation pages
    <https://gea.esac.esa.int/archive/documentation/GDR3/Data_processing/chap_cu3ast/sec_cu3ast_intro/ssec_cu3ast_intro_tansforms.html>`_.

    Attributes
    ----------
    mastorad : float
        Numerical factor to convert milliarcseconds ro radians.
    """

    def __init__(self):
        """
        Class constructor/initializer.
        """
        self.mastorad = np.pi / (180 * 3600 * 1000)

    def propagate_astrometry(
        self, phi, theta, parallax, muphistar, mutheta, vrad, t0, t1
    ):
        """
        Propagate the astrometric parameters of a source from the reference epoch t0 to
        the new epoch t1.

        Parameters
        ----------
        phi : float
            Longitude at reference epoch (radians).
        theta : float
            Latitude at reference epoch (radians).
        parallax : float
            Parallax at the reference epoch (mas).
        muphistar : float
            Proper motion in longitude (including np.cos(latitude) term) at reference
            epoch (mas/yr).
        mutheta : float
            Proper motion in latitude at reference epoch (mas/yr).
        vrad : float
            Radial velocity at reference epoch (km/s). Can be set to 0 km/s if not known.
        t0 : float
            Reference epoch (Julian years).
        t1 : float
            New epoch (Julian years).

        Returns
        -------
        phi1, theta1, parallax1, muphistar1, mutheta1, murad1 : float or array
            Astrometric parameters, including the "radial proper motion" (NOT the radial
            velocity), at the new epoch.
        """

        t = t1 - t0
        p0, q0, r0 = normal_triad(phi, theta)

        # Convert input data to units of radians and Julian year. Use ICRS coordinate
        # names internally to avoid errors in translating the formulae to code.
        pmra0 = muphistar * self.mastorad
        pmdec0 = mutheta * self.mastorad
        pmr0 = vrad * parallax / au_km_year_per_sec * self.mastorad
        pmtot0sqr = (muphistar**2 + mutheta**2) * self.mastorad**2

        # Proper motion vector
        pmvec0 = pmra0 * p0 + pmdec0 * q0

        f = (1 + 2 * pmr0 * t + (pmtot0sqr + pmr0**2) * t**2) ** (-0.5)
        u = (r0 * (1 + pmr0 * t) + pmvec0 * t) * f

        _, phi1, theta1 = cartesian_to_spherical(u[0], u[1], u[2])
        parallax1 = parallax * f
        pmr1 = (pmr0 + (pmtot0sqr + pmr0**2) * t) * f**2
        pmvec1 = (pmvec0 * (1 + pmr0 * t) - r0 * pmr0**2 * t) * f**3
        p1, q1, r1 = normal_triad(phi1, theta1)
        muphistar1 = np.sum(p1 * pmvec1 / self.mastorad, axis=0)
        mutheta1 = np.sum(q1 * pmvec1 / self.mastorad, axis=0)
        murad1 = pmr1 / self.mastorad

        return phi1, theta1, parallax1, muphistar1, mutheta1, murad1

    def propagate_pos(self, phi, theta, parallax, muphistar, mutheta, vrad, t0, t1):
        """
        Propagate the position of a source from the reference epoch t0 to the new epoch
        t1.

        Parameters
        ----------
        phi : float
            Longitude at reference epoch (radians).
        theta : float
            Latitude at reference epoch (radians).
        parallax : float
            Parallax at the reference epoch (mas).
        muphistar : float
            Proper motion in longitude (including np.cos(latitude) term) at reference
            epoch (mas/yr).
        mutheta : float
            Proper motion in latitude at reference epoch (mas/yr).
        vrad : float
            Radial velocity at reference epoch (km/s). Can be set to 0 km/s if not known.
        t0 : float
            Reference epoch (Julian years).
        t1 : float
            New epoch (Julian years).

        Returns
        -------
        phi, theta : float
            Coordinates the new epoch (in radians)
        """
        (phi1, theta1, _, _, _, _,) = self.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t1
        )
        return phi1, theta1

    def propagate_astrometry_and_covariance_matrix(self, a0, c0, t0, t1):
        """
        Propagate the astrometric parameters iand radial proper motion, as well as the
        corresponding covariance matrix, from epoch t0 to epoch t1.

        Code based on the Hipparcos Fortran implementation by Lennart Lindegren.

        Parameters
        ----------
        a0 : array_like
            6-element vector: (phi, theta, parallax, muphistar, mutheta, vrad) in units
            of (radians, radians, mas, mas/yr, mas/yr, km/s). Shape of a should be (6,)
            or (6,N), with N the number of sources for which the astrometric parameters
            are provided. The value of vrad can be set to 0 km/s if the radial velocity is not know. An appropriate uncertainty should then be provided (see below).
        c0 : array_like
            Covariance matrix stored in a 6x6 element array. This can be constructed
            from the columns listed in the Gaia catalogue. The units are [mas^2,
            mas^2/yr, mas^2/yr^2] for the various elements. Note that the elements in
            the 6th row and column should be:
            c[6,i] = c[i,6] = c[i,3] * vrad / auKmYearPerSec
            for i = 1,..,5 and
            c[6,6] = c[3,3] * (vrad^2+vrad_error^2) / auKmYearPerSec^2 +
            (parallax*vrad_error/auKmYearPerSec)^2
            The shape of c0 should be (6,6) or (N,6,6). If the radial velocity is not know the uncertainty on the radial velocity (vrad_error) should be set to the velocity dispersion of the population the source is drawn from.
        t0 : float
            Reference epoch (Julian years).
        t1 : float
            New epoch (Julian years).

        Returns
        -------
        a, c : array
            Astrometric parameters, including the "radial proper motion" (NOT the radial
            velocity), and the covariance matrix at the new epoch as a 2D matrix with
            the new variances on the diagonal and the covariance in the off-diagonal
            elements.
        """

        zero, one, two, three = 0, 1, 2, 3
        tau = t1 - t0

        # Calculate the normal triad [p0 q0 r0] at t0
        p0, q0, r0 = normal_triad(a0[0], a0[1])

        # Convert to internal units (radians, Julian year)
        par0 = a0[2] * self.mastorad
        pma0 = a0[3] * self.mastorad
        pmd0 = a0[4] * self.mastorad
        pmr0 = a0[5] * a0[2] / au_km_year_per_sec * self.mastorad

        # Proper motion vector
        pmvec0 = pma0 * p0 + pmd0 * q0

        # Auxiliary quantities
        tau2 = tau * tau
        pm02 = pma0**2 + pmd0**2
        w = one + pmr0 * tau
        f2 = one / (one + two * pmr0 * tau + (pm02 + pmr0**2) * tau2)
        f = np.sqrt(f2)
        f3 = f2 * f
        f4 = f2 * f2

        # Position vector and parallax at t1
        u = (r0 * w + pmvec0 * tau) * f
        _, ra, dec = cartesian_to_spherical(u[0], u[1], u[2])
        par = par0 * f

        # Proper motion vector and radial proper motion at t1
        pmvec = (pmvec0 * (one + pmr0 * tau) - r0 * pmr0**2 * tau) * f3
        pmr = (pmr0 + (pm02 + pmr0**2) * tau) * f2

        # Normal triad at t1
        p, q, r = normal_triad(ra, dec)

        # Convert parameters at t1 to external units (mas, Julian year)
        pma = np.sum(p * pmvec, axis=0)
        pmd = np.sum(q * pmvec, axis=0)

        a = np.zeros_like(a0)
        a[0] = ra
        a[1] = dec
        a[2] = par / self.mastorad
        a[3] = pma / self.mastorad
        a[4] = pmd / self.mastorad
        a[5] = pmr / self.mastorad

        # Auxiliary quantities for the partial derivatives

        pmz = pmvec0 * f - three * pmvec * w
        pp0 = np.sum(p * p0, axis=0)
        pq0 = np.sum(p * q0, axis=0)
        pr0 = np.sum(p * r0, axis=0)
        qp0 = np.sum(q * p0, axis=0)
        qq0 = np.sum(q * q0, axis=0)
        qr0 = np.sum(q * r0, axis=0)
        ppmz = np.sum(p * pmz, axis=0)
        qpmz = np.sum(q * pmz, axis=0)

        jacobian = np.zeros_like(c0)
        if c0.ndim == 2:
            jacobian = jacobian[np.newaxis, :, :]

        # Partial derivatives
        jacobian[:, 0, 0] = pp0 * w * f - pr0 * pma0 * tau * f
        jacobian[:, 0, 1] = pq0 * w * f - pr0 * pmd0 * tau * f
        jacobian[:, 0, 2] = zero
        jacobian[:, 0, 3] = pp0 * tau * f
        jacobian[:, 0, 4] = pq0 * tau * f
        jacobian[:, 0, 5] = -pma * tau2

        jacobian[:, 1, 0] = qp0 * w * f - qr0 * pma0 * tau * f
        jacobian[:, 1, 1] = qq0 * w * f - qr0 * pmd0 * tau * f
        jacobian[:, 1, 2] = zero
        jacobian[:, 1, 3] = qp0 * tau * f
        jacobian[:, 1, 4] = qq0 * tau * f
        jacobian[:, 1, 5] = -pmd * tau2

        jacobian[:, 2, 0] = zero
        jacobian[:, 2, 1] = zero
        jacobian[:, 2, 2] = f
        jacobian[:, 2, 3] = -par * pma0 * tau2 * f2
        jacobian[:, 2, 4] = -par * pmd0 * tau2 * f2
        jacobian[:, 2, 5] = -par * w * tau * f2

        jacobian[:, 3, 0] = -pp0 * pm02 * tau * f3 - pr0 * pma0 * w * f3
        jacobian[:, 3, 1] = -pq0 * pm02 * tau * f3 - pr0 * pmd0 * w * f3
        jacobian[:, 3, 2] = zero
        jacobian[:, 3, 3] = (
            pp0 * w * f3 - two * pr0 * pma0 * tau * f3 - three * pma * pma0 * tau2 * f2
        )
        jacobian[:, 3, 4] = (
            pq0 * w * f3 - two * pr0 * pmd0 * tau * f3 - three * pma * pmd0 * tau2 * f2
        )
        jacobian[:, 3, 5] = ppmz * tau * f2

        jacobian[:, 4, 0] = -qp0 * pm02 * tau * f3 - qr0 * pma0 * w * f3
        jacobian[:, 4, 1] = -qq0 * pm02 * tau * f3 - qr0 * pmd0 * w * f3
        jacobian[:, 4, 2] = zero
        jacobian[:, 4, 3] = (
            qp0 * w * f3 - two * qr0 * pma0 * tau * f3 - three * pmd * pma0 * tau2 * f2
        )
        jacobian[:, 4, 4] = (
            qq0 * w * f3 - two * qr0 * pmd0 * tau * f3 - three * pmd * pmd0 * tau2 * f2
        )
        jacobian[:, 4, 5] = qpmz * tau * f2

        jacobian[:, 5, 0] = zero
        jacobian[:, 5, 1] = zero
        jacobian[:, 5, 2] = zero
        jacobian[:, 5, 3] = two * pma0 * w * tau * f4
        jacobian[:, 5, 4] = two * pmd0 * w * tau * f4
        jacobian[:, 5, 5] = (w**2 - pm02 * tau2) * f4

        jacobian_transposed = np.zeros_like(jacobian)
        for i in range(jacobian.shape[0]):
            jacobian_transposed[i] = jacobian[i].T

        if c0.ndim == 2:
            c = np.matmul(
                jacobian, np.matmul(c0[np.newaxis, :, :], jacobian_transposed)
            )
        else:
            c = np.matmul(jacobian, np.matmul(c0, jacobian_transposed))

        return a, np.squeeze(c)



"""
Gaia photometry tools.

This package provides classes and functions for transforming between the Gaia
photometric system and other photometric systems. The formulae and tables are
taken from the `Gaia science performance pages
<https://www.cosmos.esa.int/web/gaia/science-performance>`_ and from the
paper by `Jordi et al <http://adsabs.harvard.edu/abs/2010A%26A...523A..48J>`_.
"""

from . import transformations
from . import utils



"""
Unit tests for the utils module.
"""

import numpy as np

from pygaia.utils import construct_covariance_matrix
from pygaia.astrometry.constants import au_km_year_per_sec


class test_utils(np.testing.TestCase):
    def test_construct_covariance_matrix_onesource(self):
        """
        Check that the covariance matrix is constructed correctly from the inputs
        provided.
        """

        cvec = np.arange(15) + 1
        parallax = 16
        vrad = 17 * au_km_year_per_sec
        vrad_error = 18 * au_km_year_per_sec

        cmat = construct_covariance_matrix(cvec, parallax, vrad, vrad_error)

        self.assertEquals(6, cmat.shape[0])
        self.assertEquals(6, cmat.shape[1])

        iu = np.triu_indices(6, k=1)

        np.testing.assert_equal(cmat[iu], cmat[iu[1], iu[0]])
        np.testing.assert_equal((np.arange(5) + 1) ** 2, np.diag(cmat)[0:-1])

        expected = np.array([12, 21, 32, 45, 60, 88, 120, 156, 210, 300])
        iuu = np.triu_indices(5, k=1)
        np.testing.assert_equal(expected, cmat[:-1, :-1][iuu])

        for i in range(5):
            np.testing.assert_almost_equal(cmat[i, 5], cmat[i, 2] * 17, 1)

        np.testing.assert_almost_equal(
            cmat[5, 5], 9 * (17**2 + 18**2) + (16 * 18) ** 2, 1
        )

    def test_construct_covariance_matrix_nsources(self):
        """
        Check that the covariance matrix is constructed correctly from the inputs
        provided.
        """

        cvec = np.zeros((2, 15))
        cvec[0] = np.arange(15) + 1
        cvec[1] = -cvec[0]
        parallax = np.array([16, -16])
        vrad = np.array([17, -17]) * au_km_year_per_sec
        vrad_error = np.array([18, -18]) * au_km_year_per_sec

        cmat = construct_covariance_matrix(cvec, parallax, vrad, vrad_error)

        self.assertEquals(2, cmat.shape[0])
        self.assertEquals(6, cmat.shape[1])
        self.assertEquals(6, cmat.shape[2])

        iu = np.triu_indices(6, k=1)

        np.testing.assert_equal(cmat[0, iu[0], iu[1]], cmat[0, iu[1], iu[0]])
        np.testing.assert_equal((np.arange(5) + 1) ** 2, np.diag(cmat[0])[0:-1])

        expected = np.array([12, 21, 32, 45, 60, 88, 120, 156, 210, 300])
        iuu = np.triu_indices(5, k=1)
        np.testing.assert_equal(expected, cmat[0, :-1, :-1][iuu])

        for i in range(5):
            np.testing.assert_almost_equal(cmat[0, i, 5], cmat[0, i, 2] * 17, 1)

        np.testing.assert_almost_equal(
            cmat[0, 5, 5], 9 * (17**2 + 18**2) + (16 * 18) ** 2, 1
        )

        np.testing.assert_equal(cmat[1, iu[0], iu[1]], cmat[1, iu[1], iu[0]])
        np.testing.assert_equal((np.arange(5) + 1) ** 2, np.diag(cmat[1])[0:-1])

        expected = -np.array([12, 21, 32, 45, 60, 88, 120, 156, 210, 300])
        iuu = np.triu_indices(5, k=1)
        np.testing.assert_equal(expected, cmat[1, :-1, :-1][iuu])

        for i in range(5):
            np.testing.assert_almost_equal(cmat[1, i, 5], cmat[1, i, 2] * (-17), 1)

        np.testing.assert_almost_equal(
            cmat[1, 5, 5], 9 * (17**2 + 18**2) + (16 * 18) ** 2, 1
        )



r"""
Provides functions for simulating the photometric (:math:`G`, :math:`G_\mathrm{BP}`,
:math:`G_\mathrm{RP}`) uncertainties on the Gaia broad-band photometry for Gaia DR3+.
The code reproduces the uncertainty model described on the `Gaia science performance
<https://www.cosmos.esa.int/web/gaia/science-performance#photometric%20performance>`_
pages.

Code taken (with permission) from the notebook at
https://github.com/gaia-dpci/gaia-dr3-photometric-uncertainties
"""
import os
import pandas as pd
import numpy as np
import scipy.interpolate as interpolate

__all__ = ["LogMagUncertainty", "magnitude_uncertainty"]

_ROOT = os.path.abspath(os.path.dirname(__file__))
_spline_csv_file = os.path.join(_ROOT, "data", "LogErrVsMagSpline2.csv")
_default_release = "dr4"
_nobs_drs_bands = {
    "dr3": {"g": 351, "bp": 40, "rp": 40},
    "dr4": {"g": 620, "bp": 70, "rp": 70},
    "dr5": {"g": 1240, "bp": 140, "rp": 140},
}


class LogMagUncertainty:
    r"""
    Estimate the log(mag) vs mag uncertainty for :math:`G`, :math:`G_\mathrm{BP}`,
    :math:`G_\mathrm{RP}` based on Gaia EDR3 photometry.

    The code in this class is a modified version of the Edr3LogMagUncertainty code in
    the Jypyter notebook EDR3_Photometric_Uncertainties.ipynb at
    https://github.com/gaia-dpci/gaia-dr3-photometric-uncertainties. The additional function estimate_for_maglist() is included to facilitate calculating the uncertainties for the number of observationis corresponding to a given Gaia data release and for the input list of magnitudes.
    """

    def __init__(self):
        """ """
        _df = pd.read_csv(_spline_csv_file)
        splines = dict()
        splines["g"] = self.__init_spline(_df, "knots_G", "coeff_G")
        splines["bp"] = self.__init_spline(_df, "knots_BP", "coeff_BP")
        splines["rp"] = self.__init_spline(_df, "knots_RP", "coeff_RP")
        self.__splines = splines
        self.__nobs = {"g": 200, "bp": 20, "rp": 20}
        self.__nobs_drs = _nobs_drs_bands

    def estimate(
        self, band, nobs: np.array([], int) = 0, mag_range=None, mag_samples=1000
    ):
        """
        Estimate the log(mag) vs mag uncertainty

        Parameters
        ----------
        band : str
            name of the band for which the uncertainties should be estimated (case-insensitive). Must be one of "g", "gbp", or "grp".
        nobs : ndarray, int
            number of observations for which the uncertainties should be estimated.
            Must be a scalar integer value or an array of integer values.
        mag_range : array_like
            Magnitude range over which the spline should be evaluated.
            The default and maximum valid range is (4, 21)
        mag_samples : int
            Number evenly spaced magnitudes (over the mag_range interval) at which the
            splines will be estimated. Default: 1000

        Raises
        ------
        ValueError
            For wrong inputs.

        Returns
        -------
        df : DataFrame
            Pandas dataframe with the interpolated log(mag) uncertainty vs mag.
            The magnitude column is named mag_g, mag_bp, or mag_rp depending of the requested band.
            A column for each value of nobs is provided, in the default case the column is logU_200.
        """
        band = band.lower()
        if band not in ["g", "bp", "rp"]:
            raise ValueError(f"Unknown band: {band}")
        if mag_range is None:
            mag_range = (4.0, 21.0)
        else:
            if mag_range[0] < 4.0:
                raise ValueError(
                    f"Uncertainties can be estimated only in the range {band}[4, 21]"
                )
            elif mag_range[1] > 21.0:
                raise ValueError(
                    f"Uncertainties can be estimated only in the range {band}[4, 21]"
                )
            elif mag_range[0] > mag_range[1]:
                raise ValueError("Malformed magnitude range")
        #
        xx = np.linspace(mag_range[0], mag_range[1], mag_samples)
        __cols = self.__compute_nobs(band, xx, nobs)
        __dc = {f"mag_{band}": xx, **__cols}
        return pd.DataFrame(data=__dc)

    def estimate_for_maglist(
        self, band, maglist: np.array([], float) = 15.0, release=_default_release
    ):
        """
        Estimate the log(mag) vs mag uncertainty

        Parameters
        ----------
        band : str
            name of the band for which the uncertainties should be estimated (case-insensitive). Must be one of "g", "gbp", or "grp".
        maglist : ndarray, float
            List of magnitudes (corresponding to the requested band) for which the
            uncertainties should be estimated. Must be a scalar float value or an array
            of float values. The values must be in the range [4, 21].
        release : str
            Gaia data release for which the uncertainties are simulated (case-insensitive). Must be one of "dr3", "dr4", or "dr5".

        Raises
        ------
        ValueError
            For wrong inputs.

        Returns
        -------
        df : DataFrame
            Pandas dataframe with the interpolated log(mag) uncertainty vs mag.
            The magnitude column is named mag_g, mag_bp, or mag_rp depending of the requested band.
            A column for each value of nobs is provided, in the default case the column is logU_200.
        """
        band = band.lower()
        release = release.lower()
        if band not in ["g", "bp", "rp"]:
            raise ValueError(f"Unknown band: {band}")
        if release not in ["dr3", "dr4", "dr5"]:
            raise ValueError(f"Unknown data release: {release}")
        if np.any(maglist < 4.0) or np.any(maglist > 21.0):
            raise ValueError(
                f"One or more of the values in maglist is outside the range [4, 21]"
            )
        #
        __cols = self.__compute_nobs(band, maglist, self.__nobs_drs[release][band])
        __dc = {f"mag_{band}": maglist, **__cols}
        return pd.DataFrame(data=__dc)

    def __init_spline(self, df, col_knots, col_coeff):
        __ddff = df[[col_knots, col_coeff]].dropna()
        return interpolate.BSpline(
            __ddff[col_knots], __ddff[col_coeff], 3, extrapolate=False
        )

    def __compute_nobs(self, band, xx, nobs):
        if isinstance(nobs, int):
            nobs = [nobs]
        __out = dict()
        for num in nobs:
            if num < 0:
                raise ValueError(f"Number of observations should be strictly positive")
            if num == 0:
                __out[f"logU_{self.__nobs[band]:d}"] = self.__splines[band](xx)
            else:
                __out[f"logU_{num:d}"] = self.__splines[band](xx) - np.log10(
                    np.sqrt(num) / np.sqrt(self.__nobs[band])
                )
        return __out


def magnitude_uncertainty(
    band, maglist: np.array([], float) = 15.0, release=_default_release
):
    r"""
    Provide the uncertainty for :math:`G`, :math:`G_\mathrm{BP}`, and
    :math:`G_\mathrm{RP}` as a function of magnitude (math:`G`, :math:`G_\mathrm{BP}`,
    and :math:`G_\mathrm{RP}`, respectively).

    Parameters
    ----------
    band : str
        name of the band for which the uncertainties are requested (case-insensitive). Must be one of "g", "gbp", or "grp".
    maglist : ndarray, float
        List of magnitudes in the same band for which the uncertainties are requested.
        Must be a scalar float value or an array of float values. The values must be in
        the range [4, 21].
    release : str
        Gaia data release for which the uncertainties are requested (case-insensitive). Must be one of "dr3", "dr4", or "dr5".

    Returns
    -------
    uncs : ndarray, float
        Array with uncertainties in units of mmag.
    """
    lmu = LogMagUncertainty()
    logunc = lmu.estimate_for_maglist(band, maglist, release)[
        f"logU_{_nobs_drs_bands[release.lower()][band.lower()]}"
    ].to_numpy()
    return np.power(10.0, logunc) * 1000



"""
Provides functions for simulating the radial velocity uncertainties in the Gaia DR3+
catalogue data. The code reproduces the uncertainty model described on the Gaia `science
performance pages
<https://www.cosmos.esa.int/web/gaia/science-performance#spectroscopic%20performance>`_.
"""
import numpy as np
import ujson
import os

__all__ = ["radial_velocity_uncertainty"]

_default_release = "dr4"

_ROOT = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(_ROOT, "data", "rv_uncertainty_model_coeffs.json")) as fp:
    _rv_unc_model = ujson.load(fp)
    fp.close()

_rv_nb_transits_dr4 = 32
_rv_nb_transits_dr5 = 64
_grvs_zeropoint = 21.317
_exposure_time = 4.4167032
_n_spectrum_per_transit = 3
_pixel_width_al = 0.02453
_band_width = 24.0
_median_background = 4.7
_n_ac_samples_bright = 10
_n_ac_samples_faint = 1
_ron = 3.2


def _in_interval(a: np.array([], float), left, right, closed="both"):
    """
    Check whether the input number or array elements lie within the given interval.

    Parameters
    ----------
    a : ndarray, float
        Input number(s) to check.
    left : float
        Left bound of interval, where left<=right.
    right : float
        Right bound of interval, where right>=left.
    closed : str
        Can be 'both' for a closed interval, 'neither' for an open interval, 'left' for a left-closed interval, or 'right' for a right-closed interval.

    Returns
    -------
    result : ndarray, boolean
        Result of the check, True or False
    """
    if left > right:
        raise ValueError("Left must be less than or equal to right")
    if closed == "both":
        return (a >= left) & (a <= right)
    elif closed == "left":
        return (a >= left) & (a < right)
    elif closed == "right":
        return (a > left) & (a <= right)
    elif closed == "neither":
        return (a > left) & (a < right)
    else:
        raise ValueError("The closed parameter must be one of both|left|right|neither")


def radial_velocity_uncertainty(
    grvs: np.array([], float),
    teff: np.array([], float),
    logg: np.array([], float),
    release=_default_release,
):
    r"""
    Simulate the Gaia DR3 radial velocity uncertainty for the input list of
    :math:`G_\mathrm{RVS}`, :math:`T_\mathrm{eff}`, and :math:`\log(g)` values.

    Parameters
    ----------
    grvs : ndarray, float
        Value(s) of :math:`G_\mathrm{RVS}` for which the calculate the radial velocity uncertainty.
    teff : ndarray, float
        Value(s) of :math:`T_\mathrm{eff}` (in K) for which to calculate the radial velocity uncertainty.
    logg : ndarray, float
        Value(s) of :math:`\log(g)` for which to calculate the radial velocity uncertainty.

    Returns
    -------
    sigma_rv : ndarray, float
        Value(s) of the radial velocity uncertainty. NaNs are returned for input
        magnitudes, temperatures, and surface gravities outside the grids as defined in
        `Katz et al (2022)
        <https://ui.adsabs.harvard.edu/abs/2022arXiv220605902K/abstracti>`_ (their
        figures E.1 and F.1), or outside the model validity ranges quoted on the `Gaia
        Science Performance pages
        <https://www.cosmos.esa.int/web/gaia/science-performance#spectroscopic%20performance>`_.
    """
    ggrvs = np.array(grvs)
    tteff = np.array(teff)
    llogg = np.array(logg)
    if not (
        (tteff.size == llogg.size)
        and (tteff.size == ggrvs.size)
        and (llogg.size == ggrvs.size)
    ):
        raise ValueError("Arrays grvs, teff. and logg must be of same size")
    rv_uncs = np.full(tteff.shape, np.nan)
    if release.upper() == "DR3":
        coeffs = "dr3coeffs"
    else:
        coeffs = "dr45coeffs"
        if release.upper() == "DR4":
            rv_nb_transits = _rv_nb_transits_dr4
        else:
            rv_nb_transits = _rv_nb_transits_dr5

    for group in _rv_unc_model.keys():
        slots = _in_interval(
            tteff,
            _rv_unc_model[group]["teff"][0],
            _rv_unc_model[group]["teff"][1],
            _rv_unc_model[group]["teff"][2],
        ) & _in_interval(
            llogg,
            _rv_unc_model[group]["logg"][0],
            _rv_unc_model[group]["logg"][1],
            _rv_unc_model[group]["logg"][2],
        )
        if np.any(slots):
            if release.upper() == "DR3":
                rv_uncs[slots] = _rv_unc_model[group][coeffs]["sfloor"] + _rv_unc_model[
                    group
                ][coeffs]["b"] * np.exp(
                    _rv_unc_model[group][coeffs]["a"]
                    * (ggrvs[slots] - _rv_unc_model[group][coeffs]["grvs0"])
                )
            else:
                rv_expected_sig_to_noise = np.zeros_like(ggrvs)
                collected_signal = (
                    np.power(10.0, 0.4 * (_grvs_zeropoint - ggrvs))
                    * _exposure_time
                    * _n_spectrum_per_transit
                    * rv_nb_transits
                    * (_pixel_width_al / _band_width)
                )
                bck_per_sample = (
                    _median_background * _n_spectrum_per_transit * rv_nb_transits
                )
                rn_per_sample = _ron * _ron * _n_spectrum_per_transit * rv_nb_transits
                bright = ggrvs <= 7
                faint = np.logical_not(bright)
                rv_expected_sig_to_noise[bright] = collected_signal[bright] / np.sqrt(
                    collected_signal[bright]
                    + bck_per_sample * _n_ac_samples_bright
                    + rn_per_sample * _n_ac_samples_bright
                )
                rv_expected_sig_to_noise[faint] = collected_signal[faint] / np.sqrt(
                    collected_signal[faint]
                    + bck_per_sample * _n_ac_samples_faint
                    + rn_per_sample * _n_ac_samples_faint
                )
                slowsnr = _rv_unc_model[group][coeffs]["sbreak"] * np.power(
                    rv_expected_sig_to_noise[slots]
                    / _rv_unc_model[group][coeffs]["snrbreak"],
                    _rv_unc_model[group][coeffs]["f"],
                )
                shighsnr = _rv_unc_model[group][coeffs]["sfloor"] + (
                    _rv_unc_model[group][coeffs]["sbreak"]
                    - _rv_unc_model[group][coeffs]["sfloor"]
                ) * np.exp(
                    _rv_unc_model[group][coeffs]["g"]
                    * (
                        np.log10(rv_expected_sig_to_noise[slots])
                        - np.log10(_rv_unc_model[group][coeffs]["snrbreak"])
                    )
                )
                h = (
                    1
                    + np.tanh(
                        _rv_unc_model[group][coeffs]["k"]
                        * (
                            np.log10(rv_expected_sig_to_noise[slots])
                            - np.log10(_rv_unc_model[group][coeffs]["snrbreak"])
                        )
                    )
                ) / 2
                rv_uncs[slots] = h * shighsnr + (1 - h) * slowsnr

    if release.upper() == "DR3":
        rv_uncs[(ggrvs > 14) | (rv_uncs > 20.0)] = np.nan
    else:
        rv_uncs[(ggrvs > 16)] = np.nan
        rv_uncs[(ggrvs > 12) & (tteff > 7000)] = np.nan

    return rv_uncs



"""
Function for visualizing coordinate transformations on the sky.
"""
import matplotlib.pyplot as plt
import numpy as np

from pygaia.astrometry.coordinates import CoordinateTransformation

__all__ = ["plot_coordinate_transformation_on_sky"]


def _order_points_for_sky_plot(x, y):
    """
    This code takes care of ordering the points (x,y), calculated for a sky map parallel
    or merdian, such that the drawing code can start at one end of the curve and end at
    the other (so no artifacts due to connecting the disjoint ends occur).

    Parameters
    ----------
    x : float array
        Set of x coordinates
    y : float array
        Set of y coordinates

    Returns
    -------
    x, y : float arrays
        Ordered set of (x,y) points
    """
    xroll = np.roll(x, 1)
    yroll = np.roll(y, 1)
    distance = (xroll - x) ** 2 + (yroll - y) ** 2
    indexmax = np.argmax(distance)
    return np.roll(x, -indexmax), np.roll(y, -indexmax)


def plot_coordinate_transformation_on_sky(
    transformation,
    fig,
    outfile=None,
    notitle=False,
    nolabels=False,
    lc=plt.cm.get_cmap("tab10").colors[0],
    tc=plt.cm.get_cmap("tab10").colors[1],
    lonpos=True,
    skyproj="hammer",
):
    """
    Produce a sky-plot in a given coordinate system with the meridians and parallels for
    another coordinate system overlayed. The coordinate systems are specified through
    the pygaia.coordinates.Transformations enum. For example for Transformations.GAL2ECL
    the sky plot will be in Ecliptic coordinates with the Galactic coordinate grid
    overlayed.

    Parameters
    --------
    transformation : pygaia.astrometry.coordinates.Transformations instance
        The coordinate transformation for which to make the plot (e.g.,
        Transformations.GAL2ECL).
    fig : matplotlib.figure.Figure
        Empty Figure instance in which to create the plot.
    outfile: str
        Save plot to this output file. Make sure an extension (.pdf, .png, etc) is included.
    notitle : boolean
        If true do not include the plot title.
    nolabels : boolean
        If true do not include plot labels.
    inax : matplotlib.axes.Axes
        If provided use the input Axes instance for plotting.
    lc : Matplotlib color specification
        Colour for gridlines.
    tc : Matplotlib color specification
        Colour for text labels.
    lonpos : boolean
        If true use longitude labels between 0 and 360 degrees.
    skyproj : str
        The geographic projection to use. Must be one of "aitoff", "hammer", or "mollweide". Default is "hammer".

    Returns
    -------

    Nothing.
    """
    if not skyproj.lower() in ["hammer", "mollweide", "aitoff"]:
        raise ValueError("Only Aitoff, Hammer, or Mollweide projections are supported")
    ct = CoordinateTransformation(transformation)

    parallels = np.arange(-80.0, 90.0, 10.0)
    meridians = np.arange(0.0, 375.0, 15.0)
    meridian_max = np.deg2rad(85.0)

    addtolabel = 0
    if lonpos:
        addtolabel = 360

    ax = fig.add_subplot(projection=skyproj)
    ax.set_longitude_grid(360)
    ax.set_latitude_grid(180)

    for thetaDeg in parallels:
        if np.mod(thetaDeg, 30) == 0:
            a = 1
        else:
            a = 0.3
        phi = np.linspace(-np.pi, np.pi, 1001)
        theta = np.zeros_like(phi) + np.deg2rad(thetaDeg)
        phirot, thetarot = ct.transform_sky_coordinates(phi, theta)
        phirot[(phirot > np.pi)] = phirot[(phirot > np.pi)] - 2 * np.pi

        indices = phirot >= 0.0
        xplot = phirot[indices]
        yplot = thetarot[indices]
        if any(indices):
            xplot, yplot = _order_points_for_sky_plot(xplot, yplot)
            ax.plot(-xplot, yplot, "-", color=lc, alpha=a)

        indices = phirot < 0.0
        xplot = phirot[indices]
        yplot = thetarot[indices]
        if any(indices):
            xplot, yplot = _order_points_for_sky_plot(xplot, yplot)
            ax.plot(-xplot, yplot, "-", color=lc, alpha=a)

    for phiDeg in meridians:
        if np.mod(phiDeg, 30) == 0:
            a = 1
        else:
            a = 0.3
        theta = np.linspace(-meridian_max, meridian_max, 1001)
        phi = np.zeros_like(theta) + np.deg2rad(phiDeg)
        phirot, thetarot = ct.transform_sky_coordinates(phi, theta)
        phirot[(phirot > np.pi)] = phirot[(phirot > np.pi)] - 2 * np.pi

        indices = phirot >= 0.0
        xplot = phirot[indices]
        yplot = thetarot[indices]
        if any(indices):
            xplot, yplot = _order_points_for_sky_plot(xplot, yplot)
            ax.plot(-xplot, yplot, "-", color=lc, alpha=a)

        indices = phirot < 0.0
        xplot = phirot[indices]
        yplot = thetarot[indices]
        if any(indices):
            xplot, yplot = _order_points_for_sky_plot(xplot, yplot)
            ax.plot(-xplot, yplot, "-", color=lc, alpha=a)

    if not notitle:
        plt.title(
            f"{skyproj.capitalize()} projection in "
            + ct.target_coordinates()
            + " coordinates with the corresponding "
            + ct.start_coordinates()
            + " grid overlayed"
        )

    if not nolabels:
        for theta in np.arange(-60, 90, 30):
            phirot, thetarot = ct.transform_sky_coordinates(0.0, np.deg2rad(theta))
            phirot[(phirot > np.pi)] = phirot[(phirot > np.pi)] - 2 * np.pi
            ax.text(
                -phirot,
                thetarot,
                "${0}$".format(theta),
                fontsize=16,
                va="bottom",
                ha="center",
                color=tc,
            )
        for phi in np.arange(-150, 0, 30):
            phirot, thetarot = ct.transform_sky_coordinates(np.deg2rad(phi), 0.0)
            phirot[(phirot > np.pi)] = phirot[(phirot > np.pi)] - 2 * np.pi
            ax.text(
                -phirot,
                thetarot,
                "${0}$".format(phi + addtolabel),
                fontsize=16,
                va="bottom",
                ha="center",
                color=tc,
            )
        for phi in np.arange(30, 210, 30):
            phirot, thetarot = ct.transform_sky_coordinates(np.deg2rad(phi), 0.0)
            phirot[(phirot > np.pi)] = phirot[(phirot > np.pi)] - 2 * np.pi
            ax.text(
                -phirot,
                thetarot,
                "${0}$".format(phi),
                fontsize=16,
                va="bottom",
                ha="center",
                color=tc,
            )

    if outfile is not None:
        plt.savefig(outfile)



"""
Provides vector astrometry functions.
"""
import numpy as np
from .constants import au_mas_parsec, au_km_year_per_sec

__all__ = [
    "spherical_to_cartesian",
    "cartesian_to_spherical",
    "normal_triad",
    "elementary_rotation_matrix",
    "phase_space_to_astrometry",
    "astrometry_to_phase_space",
]


def spherical_to_cartesian(r, phi, theta):
    """
    Convert spherical to Cartesian coordinates. The input can be scalars or
    1-dimensional numpy arrays. Note that the angle coordinates follow the astronomical
    convention of using elevation (declination, latitude) rather than its complement
    (pi/2-elevation), where the latter is commonly used in the mathematical treatment of
    spherical coordinates.

    Parameters
    ----------
    r : float or float array
        Length of input Cartesian vector.
    phi : float or float array
        Longitude-like angle (e.g., right ascension, ecliptic longitude) in radians
    theta : float or float array
        Latitide-like angle (e.g., declination, ecliptic latitude) in radians

    Returns
    -------
    x, y, z : float or float array
        The Cartesian vector components x, y, z
    """
    ctheta = np.cos(theta)
    x = r * np.cos(phi) * ctheta
    y = r * np.sin(phi) * ctheta
    z = r * np.sin(theta)
    return x, y, z


def cartesian_to_spherical(x, y, z):
    r"""
    Convert Cartesian to spherical coordinates. The input can be scalars or
    1-dimensional numpy arrays. Note that the angle coordinates follow the astronomical
    convention of using elevation (declination, latitude) rather than its complement
    (pi/2-elevation), which is commonly used in the mathematical treatment of spherical
    coordinates.

    Parameters
    ----------
    x : float or float array
        Cartesian vector component along the X-axis
    y : float or float array
        Cartesian vector component along the Y-axis
    z : float or float array
        Cartesian vector component along the Z-axis

    Returns
    -------
    r, phi, theta : float or float array
        The spherical coordinates r=np.sqrt(x*x+y*y+z*z), longitude phi, latitude theta.

    Raises
    ------
    ValueError
        When one of the input (x,y,z) coordinates correspond to r=0.

    Notes
    -----
        The longitude angle is between 0 and :math:`+2\pi`.
    """
    rCylSq = x * x + y * y
    r = np.sqrt(rCylSq + z * z)
    if np.any(r == 0.0):
        raise ValueError("Error: one or more of the input points is at distance zero.")
    phi = np.arctan2(y, x)
    phi = np.where(phi < 0.0, phi + 2 * np.pi, phi)
    return r, phi, np.arctan2(z, np.sqrt(rCylSq))


def normal_triad(phi, theta):
    """
    Calculate the so-called normal triad [p, q, r] associated with a spherical
    coordinate system. The three vectors are:

    p - The unit tangent vector in the direction of increasing longitudinal angle phi.
    q - The unit tangent vector in the direction of increasing latitudinal angle theta.
    r - The unit vector toward the point (phi, theta).

    Parameters
    ----------
    phi : float or array
        Longitude-like angle (e.g., right ascension, ecliptic longitude) in radians
    theta : float or array
        Latitide-like angle (e.g., declination, ecliptic latitude) in radians

    Returns
    -------
    p, q, r : array
        The normal triad as the vectors p, q, r as (N,3) arrays
    """
    sphi = np.sin(phi)
    stheta = np.sin(theta)
    cphi = np.cos(phi)
    ctheta = np.cos(theta)
    p = np.array([-sphi, cphi, np.zeros_like(phi)])
    q = np.array([-stheta * cphi, -stheta * sphi, ctheta])
    r = np.array([ctheta * cphi, ctheta * sphi, stheta])
    return p, q, r


def elementary_rotation_matrix(axis, rotationAngle):
    """
    Construct an elementary rotation matrix describing a rotation around the x, y, or
    z-axis.

    Parameters
    ----------
    axis : str
        Axis around which to rotate ("x", "X", "y", "Y", "z", or "Z")
    rotationAngle : float
        The rotation angle in radians

    Returns
    -------
    rmat : array
        The rotation matrix

    Raises
    ------
    ValueError
        If an unsupported rotation axis string is supplied.

    Examples
    --------
    >>> rotmat = elementaryRotationMatrix("y", np.pi/6.0)
    """
    if axis.upper() == "X":
        return np.array(
            [
                [1.0, 0.0, 0.0],
                [0.0, np.cos(rotationAngle), np.sin(rotationAngle)],
                [0.0, -np.sin(rotationAngle), np.cos(rotationAngle)],
            ]
        )
    elif axis.upper() == "Y":
        return np.array(
            [
                [np.cos(rotationAngle), 0.0, -np.sin(rotationAngle)],
                [0.0, 1.0, 0.0],
                [np.sin(rotationAngle), 0.0, np.cos(rotationAngle)],
            ]
        )
    elif axis.upper() == "Z":
        return np.array(
            [
                [np.cos(rotationAngle), np.sin(rotationAngle), 0.0],
                [-np.sin(rotationAngle), np.cos(rotationAngle), 0.0],
                [0.0, 0.0, 1.0],
            ]
        )
    else:
        raise ValueError("Unknown rotation axis " + axis + "!")


def phase_space_to_astrometry(x, y, z, vx, vy, vz):
    r"""
    From the given phase space coordinates calculate the astrometric observables,
    including the radial velocity, which here is seen as the sixth astrometric
    parameter. The phase space coordinates are assumed to represent barycentric (i.e.
    centred on the Sun) positions and velocities.

    Parameters
    ----------
    x : float or float array
        The x component of the barycentric position vector (in pc or kpc).
    y : float or float array
        The y component of the barycentric position vector (in pc or kpc).
    z : float or float array
        The z component of the barycentric position vector (in pc or kpc).
    vx : float or float array
        The x component of the barycentric velocity vector (in km/s).
    vy : float or float array
        The y component of the barycentric velocity vector (in km/s).
    vz : float or float array
        The z component of the barycentric velocity vector (in km/s).

    Returns
    -------
    phi : float or float array
        The longitude-like angle of the position of the source (radians).
    theta : float or float array
        The latitude-like angle of the position of the source (radians).
    parallax : float or float array
        The parallax of the source (in mas or muas, see notes)
    muphistar : float or float array
        The proper motion in the longitude-like angle, multiplied by cos(theta) (mas/yr
        or muas/yr, see notes)
    mutheta : float or float array
        The proper motion in the latitude-like angle (mas/yr or muas/yr, see notes)
    vrad : float or float array
        The radial velocity (km/s)

    Notes
    -----
    This function has no mechanism to deal with units. The velocity units are always
    assumed to be km/s, and the code is set up such that for positions in pc, the return
    units for the astrometry are radians, milliarcsec, milliarcsec/year and km/s. For
    positions in kpc the return units are: radians, microarcsec, microarcsec/year, and
    km/s.

    The doppler factor :math:`k=1/(1-v_\mathrm{rad}/c)` is not used in the calculations.
    This is not a problem for sources moving at typical velocities of Galactic stars.

    """
    u, phi, theta = cartesian_to_spherical(x, y, z)
    parallax = au_mas_parsec / u
    p, q, r = normal_triad(phi, theta)
    velocitiesArray = np.array([vx, vy, vz])
    if np.isscalar(u):
        muphistar = np.dot(p, velocitiesArray) * parallax / au_km_year_per_sec
        mutheta = np.dot(q, velocitiesArray) * parallax / au_km_year_per_sec
        vrad = np.dot(r, velocitiesArray)
    else:
        muphistar = np.sum(p * velocitiesArray, axis=0) * parallax / au_km_year_per_sec
        mutheta = np.sum(q * velocitiesArray, axis=0) * parallax / au_km_year_per_sec
        vrad = np.sum(r * velocitiesArray, axis=0)

    return phi, theta, parallax, muphistar, mutheta, vrad


def astrometry_to_phase_space(phi, theta, parallax, muphistar, mutheta, vrad):
    r"""
    From the input astrometric parameters calculate the phase space coordinates. The
    output phase space coordinates represent barycentric (i.e. centred on the Sun)
    positions and velocities.

    Parameters
    ----------
    phi : float or float array
        The longitude-like angle of the position of the source (radians).
    theta : float or float array
        The latitude-like angle of the position of the source (radians).
    parallax : float or float array
        The parallax of the source (in mas or muas, see notes)
    muphistar : float or float array
        The proper motion in the longitude-like angle, multiplied by cos(theta) (mas/yr
        or muas/yr, see notes)
    mutheta : float or float array
        The proper motion in the latitude-like angle (mas/yr or muas/yr, see notes)
    vrad : float or float array
        The radial velocity (km/s)

    Returns
    -------
    x : float or float array
        The x component of the barycentric position vector (in pc or kpc).
    y : float or float array
        The y component of the barycentric position vector (in pc or kpc).
    z : float or float array
        The z component of the barycentric position vector (in pc or kpc).
    vx : float or float array
        The x component of the barycentric velocity vector (in km/s).
    vy : float or float array
        The y component of the barycentric velocity vector (in km/s).
    vz : float or float array
        The z component of the barycentric velocity vector (in km/s).

    Raises
    ------
    ValueError
        If any of the input parallaxes is non-positive.

    Notes
    -----
    This function has no mechanism to deal with units. The code is set up such that for
    input astrometry with parallaxes and proper motions in mas and mas/yr, and radial
    velocities in km/s, the phase space coordinates are in pc and km/s. For input
    astrometry with parallaxes and proper motions in muas and muas/yr, and radial
    velocities in km/s, the phase space coordinates are in kpc and km/s. Only positive
    parallaxes are accepted, an exception is thrown if this condition is not met.

    The doppler factor :math:`k=1/(1-v_mathrm{rad}/c)` is not used in the calculations.
    This is not a problem for sources moving at typical velocities of Galactic stars.

    **This function should not be used when the parallaxes have relative uncertainties
    larger than about 20 per cent** (see http://arxiv.org/abs/1507.02105 for example).
    For astrometric data with relatively large parallax errors you should consider doing
    your analysis in the data space and use forward modelling of some kind.
    """
    if np.any(parallax <= 0.0):
        raise ValueError("One or more of the input parallaxes is non-positive")
    x, y, z = spherical_to_cartesian(au_mas_parsec / parallax, phi, theta)
    p, q, r = normal_triad(phi, theta)
    transverseMotionArray = np.array(
        [
            muphistar * au_km_year_per_sec / parallax,
            mutheta * au_km_year_per_sec / parallax,
            vrad,
        ]
    )
    if np.isscalar(parallax):
        velocityArray = np.dot(np.transpose(np.array([p, q, r])), transverseMotionArray)
        vx = velocityArray[0]
        vy = velocityArray[1]
        vz = velocityArray[2]
    else:
        vx = np.zeros_like(parallax)
        vy = np.zeros_like(parallax)
        vz = np.zeros_like(parallax)
        for i in range(parallax.size):
            velocityArray = np.dot(
                np.transpose(np.array([p[:, i], q[:, i], r[:, i]])),
                transverseMotionArray[:, i],
            )
            vx[i] = velocityArray[0]
            vy[i] = velocityArray[1]
            vz[i] = velocityArray[2]
    return x, y, z, vx, vy, vz



"""
Utility functions for PyGaia.
"""
import numpy as np

from pygaia.astrometry.constants import au_km_year_per_sec

__all__ = ["construct_covariance_matrix"]


def construct_covariance_matrix(cvec, parallax, radial_velocity, radial_velocity_error):
    """
    Take the astrometric parameter standard uncertainties and the uncertainty
    correlations as quoted in the Gaia catalogue and construct the covariance matrix.

    Parameters
    ----------
    cvec : array_like
        Array of shape (15,) (1 source) or (N,15) (N sources) for the astrometric
        parameter standard uncertainties and their correlations, as listed in the Gaia
        catalogue [ra_error, dec_error, parallax_error, pmra_error, pmdec_error,
        ra_dec_corr, ra_parallax_corr, ra_pmra_corr, ra_pmdec_corr, dec_parallax_corr,
        dec_pmra_corr, dec_pmdec_corr, parallax_pmra_corr, parallax_pmdec_corr,
        pmra_pmdec_corr]. Units are (mas^2, mas^2/yr, mas^2/yr^2).
    parallax : array_like
        Source parallax(es) array of shape (N,) (mas).
    radial_velocity : array_like
        Source radial velocity (km/s, does not have to be from Gaia RVS!) as array of
        shape (N,). If the radial velocity is not known it can be set to zero.
    radial_velocity_error : array_like
        Source radial velocity  uncertainty (km/s), array of shape (N,). If the radial
        velocity is not know this can be set to the radial velocity dispersion for the
        population the source was drawn from.

    Returns
    -------
    cmat : float array
        Covariance matrix as a 6x6 array.
    """

    if np.ndim(cvec) == 1:
        cmat = np.zeros((1, 6, 6))
        nsources = 1
        cv = np.atleast_2d(cvec)
    else:
        nsources = cvec.shape[0]
        cmat = np.zeros((nsources, 6, 6))
        cv = cvec
    for k in range(nsources):
        np.fill_diagonal(cmat[k], cv[k, 0:5] ** 2)

    iu = np.triu_indices(5, k=1)
    for k in range(10):
        i = iu[0][k]
        j = iu[1][k]
        cmat[:, i, j] = cv[:, i] * cv[:, j] * cv[:, k + 5]
        cmat[:, j, i] = cmat[:, i, j]

    for k in range(nsources):
        cmat[k, 0:5, 5] = (
            cmat[k, 0:5, 2] * np.atleast_1d(radial_velocity)[k] / au_km_year_per_sec
        )
    cmat[:, 5, 0:5] = cmat[:, 0:5, 5]
    cmat[:, 5, 5] = (
        cmat[:, 2, 2]
        * (radial_velocity**2 + radial_velocity_error**2)
        / au_km_year_per_sec**2
        + (parallax * radial_velocity_error / au_km_year_per_sec) ** 2
    )

    return np.squeeze(cmat)



"""
Provides numerical constants useful for astrometric calculations.
"""

import numpy as np

__all__ = [
    "au_in_meter",
    "au_mas_parsec",
    "julian_year_seconds",
    "au_km_year_per_sec",
    "au_mas_km_year_per_sec",
    "parsec",
]

# Astronomical Unit in meter, IAU constant and defining length
au_in_meter = 149597870700.0

# AU expressed in mas*pc or muas*kpc
au_mas_parsec = 1000.0

# Number of seconds in Julian year
julian_year_seconds = 365.25 * 86400.0

# AU expressed in km*yr/s
au_km_year_per_sec = au_in_meter / (julian_year_seconds * 1000.0)

# AU expressed in mas*km*yr/s
au_mas_km_year_per_sec = au_km_year_per_sec * 180 / np.pi * 3600 * 1000

# Definition of parsec
parsec = au_in_meter * 180 * 3600 / np.pi



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
import pygaia

# -- Project information -----------------------------------------------------

project = "PyGaia"
copyright = "2012-2023, Anthony Brown, Leiden University, Gaia Data Processing and Analysis Consortium"
author = "Anthony Brown"

# The full version, including alpha/beta/rc tags
# release = "3.0"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_automodapi.automodapi",
    "matplotlib.sphinxext.plot_directive",
    "sphinx.ext.napoleon",
]
numpydoc_show_class_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"
html_title = "PyGaia"
# Not needed for now
# html_static_path = ["_static"]
html_show_sourcelink = False
html_theme_options = {
    "path_to_docs": "docs",
    "repository_url": "https://github.com/agabrown/PyGaia",
    "repository_branch": "main",
    "launch_buttons": {
        "binderhub_url": "https://mybinder.org",
        "colab_url": "https://colab.research.google.com/",
        "notebook_interface": "jupyterlab",
    },
    "use_edit_page_button": False,
    "use_issues_button": False,
    "use_repository_button": True,
    "use_download_button": True,
}
html_baseurl = "https://pygaia.readthedocs.io/en/latest/"



"""
Unit tests for the coordinates module.
"""
import numpy as np
from numpy.random import rand
import numpy.testing as npt

from pygaia.astrometry.constants import au_km_year_per_sec
from pygaia.astrometry.coordinates import CoordinateTransformation
from pygaia.astrometry.coordinates import (
    Transformations,
    EpochPropagation,
    angular_distance,
)
from pygaia.astrometry.vectorastrometry import (
    cartesian_to_spherical,
    astrometry_to_phase_space,
    phase_space_to_astrometry,
)
from pygaia.utils import construct_covariance_matrix

rng = np.random.default_rng()


class TestCoordinates(np.testing.TestCase):
    def setUp(self):
        self.basisVectorXValues = np.array([1, 0, 0])
        self.basisVectorYValues = np.array([0, 1, 0])
        self.basisVectorZValues = np.array([0, 0, 1])

        self.expectedMatIcrsToGal = np.array(
            [[-0.05, -0.87, -0.48], [+0.49, -0.44, +0.75], [-0.87, -0.20, +0.46]]
        )
        self.expectedMatEclipticToIcrs = np.array(
            [[1.0, 0.0, 0.0], [0.0, 0.92, -0.40], [0.0, 0.40, 0.92]]
        )
        self.expectedMatGalacticToEcliptic = np.array(
            [[-0.05, 0.49, -0.87], [-0.99, -0.11, 0.00], [-0.1, 0.86, 0.50]]
        )

    def test_galacticToIcrs(self):
        """
        Verify correctness of transformations from the Galactic to ICRS coordinate
        systems.
        """
        ct = CoordinateTransformation(Transformations.GAL2ICRS)
        x, y, z = ct.transform_cartesian_coordinates(
            self.basisVectorXValues, self.basisVectorYValues, self.basisVectorZValues
        )
        npt.assert_array_almost_equal(x, self.expectedMatIcrsToGal[:, 0], decimal=2)
        npt.assert_array_almost_equal(y, self.expectedMatIcrsToGal[:, 1], decimal=2)
        npt.assert_array_almost_equal(z, self.expectedMatIcrsToGal[:, 2], decimal=2)

        r, expectedAlpha, expectedDelta = cartesian_to_spherical(
            self.expectedMatIcrsToGal[:, 0],
            self.expectedMatIcrsToGal[:, 1],
            self.expectedMatIcrsToGal[:, 2],
        )
        alpha, delta = ct.transform_sky_coordinates(
            np.array([0.0, np.pi / 2.0, 0.0]), np.array([0.0, 0.0, np.pi / 2.0])
        )
        npt.assert_array_almost_equal(expectedAlpha, alpha, decimal=1)
        npt.assert_array_almost_equal(expectedDelta, delta, decimal=1)

        alpha, delta = ct.transform_sky_coordinates(0.0, np.pi / 2.0)
        npt.assert_allclose(alpha / np.pi * 180, 192.9, atol=1.0)
        npt.assert_allclose(delta / np.pi * 180, 27.1, atol=1.0)

    def test_icrsToGalactic(self):
        """
        Verify correctness of transformations from the ICRS to Galactic coordinate
        systems.
        """
        ct = CoordinateTransformation(Transformations.ICRS2GAL)
        x, y, z = ct.transform_cartesian_coordinates(
            self.basisVectorXValues, self.basisVectorYValues, self.basisVectorZValues
        )
        npt.assert_array_almost_equal(x, self.expectedMatIcrsToGal[0, :], decimal=2)
        npt.assert_array_almost_equal(y, self.expectedMatIcrsToGal[1, :], decimal=2)
        npt.assert_array_almost_equal(z, self.expectedMatIcrsToGal[2, :], decimal=2)

        r, expectedGalon, expectedGalat = cartesian_to_spherical(
            self.expectedMatIcrsToGal[0, :],
            self.expectedMatIcrsToGal[1, :],
            self.expectedMatIcrsToGal[2, :],
        )
        galon, galat = ct.transform_sky_coordinates(
            np.array([0.0, np.pi / 2.0, 0.0]), np.array([0.0, 0.0, np.pi / 2.0])
        )
        npt.assert_array_almost_equal(expectedGalon, galon, decimal=1)
        npt.assert_array_almost_equal(expectedGalat, galat, decimal=1)

    def test_icrsToEcliptic(self):
        """
        Verify correctness of transformations from the ICRS to Ecliptic coordinate
        systems.
        """
        ct = CoordinateTransformation(Transformations.ICRS2ECL)
        x, y, z = ct.transform_cartesian_coordinates(
            self.basisVectorXValues, self.basisVectorYValues, self.basisVectorZValues
        )
        npt.assert_array_almost_equal(
            x, self.expectedMatEclipticToIcrs[:, 0], decimal=2
        )
        npt.assert_array_almost_equal(
            y, self.expectedMatEclipticToIcrs[:, 1], decimal=2
        )
        npt.assert_array_almost_equal(
            z, self.expectedMatEclipticToIcrs[:, 2], decimal=2
        )

        r, expectedLambda, expectedBeta = cartesian_to_spherical(
            self.expectedMatEclipticToIcrs[:, 0],
            self.expectedMatEclipticToIcrs[:, 1],
            self.expectedMatEclipticToIcrs[:, 2],
        )
        lambdaEcl, betaEcl = ct.transform_sky_coordinates(
            np.array([0.0, np.pi / 2.0, 0.0]), np.array([0.0, 0.0, np.pi / 2.0])
        )
        npt.assert_array_almost_equal(expectedLambda, lambdaEcl, decimal=1)
        npt.assert_array_almost_equal(expectedBeta, betaEcl, decimal=1)

        alpha = 2.0 * np.pi * rng.uniform(size=100)
        delta = -np.pi / 2.0 + np.pi * rng.uniform(size=100)
        lambdaEcl, betaEcl = ct.transform_sky_coordinates(alpha, delta)
        result = 0.9175 * np.sin(delta) - 0.3978 * np.sin(alpha) * np.cos(delta)
        npt.assert_array_almost_equal(result, np.sin(betaEcl), decimal=2)

    def test_eclipticToIcrs(self):
        """
        Verify correctness of transformations from the Ecliptic to ICRS coordinate
        systems.
        """
        ct = CoordinateTransformation(Transformations.ECL2ICRS)
        x, y, z = ct.transform_cartesian_coordinates(
            self.basisVectorXValues, self.basisVectorYValues, self.basisVectorZValues
        )
        npt.assert_array_almost_equal(
            x, self.expectedMatEclipticToIcrs[0, :], decimal=2
        )
        npt.assert_array_almost_equal(
            y, self.expectedMatEclipticToIcrs[1, :], decimal=2
        )
        npt.assert_array_almost_equal(
            z, self.expectedMatEclipticToIcrs[2, :], decimal=2
        )

        r, expectedAlpha, expectedDelta = cartesian_to_spherical(
            self.expectedMatEclipticToIcrs[0, :],
            self.expectedMatEclipticToIcrs[1, :],
            self.expectedMatEclipticToIcrs[2, :],
        )
        alpha, delta = ct.transform_sky_coordinates(
            np.array([0.0, np.pi / 2.0, 0.0]), np.array([0.0, 0.0, np.pi / 2.0])
        )
        npt.assert_array_almost_equal(expectedAlpha, alpha, decimal=1)
        npt.assert_array_almost_equal(expectedDelta, delta, decimal=1)

        lambdaEcl = 2.0 * np.pi * rng.uniform(size=100)
        betaEcl = -np.pi / 2.0 + np.pi * rng.uniform(size=100)
        alpha, delta = ct.transform_sky_coordinates(lambdaEcl, betaEcl)
        result = 0.9175 * np.sin(betaEcl) + 0.3978 * np.sin(lambdaEcl) * np.cos(betaEcl)
        npt.assert_array_almost_equal(result, np.sin(delta), decimal=2)

    def test_eclipticToGalactic(self):
        """
        Verify correctness of transformations from the Ecliptic to Galactic coordinate
        systems.
        """
        ct = CoordinateTransformation(Transformations.ECL2GAL)
        x, y, z = ct.transform_cartesian_coordinates(
            self.basisVectorXValues, self.basisVectorYValues, self.basisVectorZValues
        )
        npt.assert_array_almost_equal(
            x, self.expectedMatGalacticToEcliptic[:, 0], decimal=2
        )
        npt.assert_array_almost_equal(
            y, self.expectedMatGalacticToEcliptic[:, 1], decimal=2
        )
        npt.assert_array_almost_equal(
            z, self.expectedMatGalacticToEcliptic[:, 2], decimal=2
        )

        r, expectedGalon, expectedGalat = cartesian_to_spherical(
            self.expectedMatGalacticToEcliptic[:, 0],
            self.expectedMatGalacticToEcliptic[:, 1],
            self.expectedMatGalacticToEcliptic[:, 2],
        )
        galon, galat = ct.transform_sky_coordinates(
            np.array([0.0, np.pi / 2.0, 0.0]), np.array([0.0, 0.0, np.pi / 2.0])
        )
        npt.assert_array_almost_equal(expectedGalon, galon, decimal=1)
        npt.assert_array_almost_equal(expectedGalat, galat, decimal=1)

    def test_galacticToEcliptic(self):
        """
        Verify correctness of transformations from the Galactic to Ecliptic coordinate
        systems.
        """
        ct = CoordinateTransformation(Transformations.GAL2ECL)
        x, y, z = ct.transform_cartesian_coordinates(
            self.basisVectorXValues, self.basisVectorYValues, self.basisVectorZValues
        )
        npt.assert_array_almost_equal(
            x, self.expectedMatGalacticToEcliptic[0, :], decimal=2
        )
        npt.assert_array_almost_equal(
            y, self.expectedMatGalacticToEcliptic[1, :], decimal=2
        )
        npt.assert_array_almost_equal(
            z, self.expectedMatGalacticToEcliptic[2, :], decimal=2
        )

        r, expectedLambda, expectedBeta = cartesian_to_spherical(
            self.expectedMatGalacticToEcliptic[0, :],
            self.expectedMatGalacticToEcliptic[1, :],
            self.expectedMatGalacticToEcliptic[2, :],
        )
        lambdaEcl, betaEcl = ct.transform_sky_coordinates(
            np.array([0.0, np.pi / 2.0, 0.0]), np.array([0.0, 0.0, np.pi / 2.0])
        )
        #
        # Note the abs() in the line below is required to avoid an error when -pi and pi
        # are compared. The better solution of course is to write an npt.assert
        # function that can handle modulo 2*pi cases.
        #
        npt.assert_array_almost_equal(abs(expectedLambda), abs(lambdaEcl), decimal=1)
        npt.assert_array_almost_equal(expectedBeta, betaEcl, decimal=1)

        galon = 2.0 * np.pi * rng.uniform(size=100)
        galat = -np.pi / 2.0 + np.pi * rng.uniform(size=100)
        lambdaEcl, betaEcl = ct.transform_sky_coordinates(galon, galat)
        phi = 6.38 / 180.0 * np.pi
        result = 0.4971 * np.sin(galat) + 0.8677 * np.sin(galon - phi) * np.cos(galat)
        npt.assert_array_almost_equal(result, np.sin(betaEcl), decimal=2)

    def test_transformProperMotions(self):
        """
        Verify the correct implementation of the direct transformation of proper
        motions.
        """
        ct = CoordinateTransformation(Transformations.ICRS2GAL)
        nTests = 100
        phi = 2.0 * np.pi * rng.uniform(size=nTests)
        theta = -np.pi / 2.0 + np.pi * rng.uniform(size=nTests)
        parallax = rng.uniform(size=nTests) * 99.0 + 1.0
        muphistar = -30.0 + 60.0 * rng.uniform(size=nTests)
        mutheta = -30.0 + 60.0 * rng.uniform(size=nTests)
        vrad = -200.0 + 400.0 * rng.uniform(size=nTests)

        x, y, z, vx, vy, vz = astrometry_to_phase_space(
            phi, theta, parallax, muphistar, mutheta, vrad
        )
        xrot, yrot, zrot = ct.transform_cartesian_coordinates(x, y, z)
        vxrot, vyrot, vzrot = ct.transform_cartesian_coordinates(vx, vy, vz)
        (
            phiRot,
            thetaRot,
            parRot,
            muphistarRotExpected,
            muthetaRotExpected,
            vradRot,
        ) = phase_space_to_astrometry(xrot, yrot, zrot, vxrot, vyrot, vzrot)

        muphistarRot, muthetaRot = ct.transform_proper_motions(
            phi, theta, muphistar, mutheta
        )

        npt.assert_array_almost_equal(muphistarRotExpected, muphistarRot, decimal=2)
        npt.assert_array_almost_equal(muthetaRotExpected, muthetaRot, decimal=2)

        #
        # Test scalar version of method.
        #
        for i in range(nTests):
            muphistarRot, muthetaRot = ct.transform_proper_motions(
                phi[i], theta[i], muphistar[i], mutheta[i]
            )
            npt.assert_almost_equal(muphistarRotExpected[i], muphistarRot, decimal=2)
            npt.assert_almost_equal(muthetaRotExpected[i], muthetaRot, decimal=2)

    def test_transformSkyCoordinateErrors(self):
        """
        Verify that the transformed covariance matrix for the positions remains a
        covariance matrix.
        """
        ct = CoordinateTransformation(Transformations.ICRS2GAL)
        nTests = 100
        phi = 2.0 * np.pi * rng.uniform(size=nTests)
        theta = -np.pi / 2.0 + np.pi * rng.uniform(size=nTests)
        sigPhiStar, sigTheta, rhoPhiTheta = self._generateRandomCovarianceMatrices(
            nTests
        )

        sigPhiStarRot, sigThetaRot, rhoPhiThetaRot = ct.transform_sky_coordinate_errors(
            phi, theta, sigPhiStar, sigTheta, rho_phi_theta=rhoPhiTheta
        )
        for i in range(nTests):
            self.assertGreater(sigPhiStarRot[i], 0.0)
            self.assertGreater(sigThetaRot[i], 0.0)
            self.assertTrue(-1 <= rhoPhiThetaRot[i] <= 1)

        sigPhiStarRot, sigThetaRot, rhoPhiThetaRot = ct.transform_sky_coordinate_errors(
            phi, theta, sigPhiStar, sigTheta
        )
        for i in range(nTests):
            self.assertGreater(sigPhiStarRot[i], 0.0)
            self.assertGreater(sigThetaRot[i], 0.0)
            self.assertTrue(-1 <= rhoPhiThetaRot[i] <= 1)

        sigPhiStarRot, sigThetaRot, rhoPhiThetaRot = ct.transform_sky_coordinate_errors(
            phi, theta, sigPhiStar, sigTheta, rho_phi_theta=0.7
        )
        for i in range(nTests):
            self.assertGreater(sigPhiStarRot[i], 0.0)
            self.assertGreater(sigThetaRot[i], 0.0)
            self.assertTrue(-1 <= rhoPhiThetaRot[i] <= 1)

        for i in range(nTests):
            (
                sigPhiStarRot,
                sigThetaRot,
                rhoPhiThetaRot,
            ) = ct.transform_sky_coordinate_errors(
                phi[i], theta[i], sigPhiStar[i], sigTheta[i], rho_phi_theta=0.7
            )
            self.assertGreater(sigPhiStarRot, 0.0)
            self.assertGreater(sigThetaRot, 0.0)
            self.assertTrue(-1 <= rhoPhiThetaRot <= 1)

    def test_transformProperMotionErrors(self):
        """
        Verify that the transformed covariance matrix for the proper motions remains a
        covariance matrix.
        """
        ct = CoordinateTransformation(Transformations.ICRS2GAL)
        nTests = 100
        phi = 2.0 * np.pi * rng.uniform(size=nTests)
        theta = -np.pi / 2.0 + np.pi * rng.uniform(size=nTests)
        sigPhiStar, sigTheta, rhoPhiTheta = self._generateRandomCovarianceMatrices(
            nTests
        )

        sigPhiStarRot, sigThetaRot, rhoPhiThetaRot = ct.transform_proper_motion_errors(
            phi, theta, sigPhiStar, sigTheta, rho_muphi_mutheta=rhoPhiTheta
        )
        for i in range(nTests):
            self.assertGreater(sigPhiStarRot[i], 0.0)
            self.assertGreater(sigThetaRot[i], 0.0)
            self.assertTrue(-1 <= rhoPhiTheta[i] <= 1)

        sigPhiStarRot, sigThetaRot, rhoPhiThetaRot = ct.transform_proper_motion_errors(
            phi, theta, sigPhiStar, sigTheta
        )
        for i in range(nTests):
            self.assertGreater(sigPhiStarRot[i], 0.0)
            self.assertGreater(sigThetaRot[i], 0.0)
            self.assertTrue(-1 <= rhoPhiTheta[i] <= 1)

    def test_propagate_astrometry(self):
        """
        Verify correct working of the epoch propagation class.
        """
        radtomas = 180 * 3600 * 1000 / np.pi

        ep = EpochPropagation()
        phi = 0.0
        theta = 0.0
        parallax = 10.0
        muphistar = 1.0
        mutheta = 0.0
        vrad = 0.0
        t0 = 2015.5
        t1 = 2020.5
        t2 = 2010.0
        phi1, theta1, parallax1, muphistar1, mutheta1, pmr1 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t1
        )
        self.assertGreater(phi1, 0.0)
        self.assertEquals(theta1, 0.0)
        self.assertLess(parallax1, parallax)
        self.assertLess(muphistar1, muphistar)
        self.assertEquals(mutheta1, 0.0)
        self.assertGreater(pmr1, 0.0)

        # For propagation into the future check that in absence of radial motion the
        # parallaxes and proper motions always diminish with respect to the reference
        # epoch, while the radial proper motion always becomes positive non-zero.
        nTests = 100
        phi = 2.0 * np.pi * rng.uniform(size=nTests)
        theta = np.arcsin(-1 + 2 * rng.uniform(size=nTests))
        parallax = rng.uniform(size=nTests) * 20.0 + 1.0
        muphistar = rng.uniform(size=nTests) * 5.0 - 10.0
        mutheta = rng.uniform(size=nTests) * 5.0 - 10.0
        vrad = np.zeros_like(phi)
        phi1, theta1, parallax1, muphistar1, mutheta1, pmr1 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t1
        )
        npt.assert_array_less(parallax1, parallax)
        npt.assert_array_less(
            muphistar1**2 + mutheta1**2, muphistar**2 + mutheta**2
        )
        npt.assert_array_less(0.0, pmr1)

        # For propagation into the past check that in absence of radial motion the
        # parallaxes and proper motions always diminish with respect to the reference
        # epoch, while the radial proper motion always becomes negative non-zero.
        phi2, theta2, parallax2, muphistar2, mutheta2, pmr2 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t2
        )
        npt.assert_array_less(parallax2, parallax)
        npt.assert_array_less(
            muphistar2**2 + mutheta2**2, muphistar**2 + mutheta**2
        )
        npt.assert_array_less(pmr2, 0.0)

        # For propagation into the future check that In the absence of proper motion and
        # for positive radial velocity and positive parallax at the current epoch: the
        # parallax always decreases, the radial proper motion decreases.
        muphistar = np.zeros_like(phi)
        mutheta = np.zeros_like(phi)
        vrad = rng.uniform(size=nTests) * 50.0 + 1.0
        pmr = vrad * parallax / au_km_year_per_sec
        phi1, theta1, parallax1, muphistar1, mutheta1, pmr1 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t1
        )
        npt.assert_almost_equal(phi1, phi, 12)
        npt.assert_almost_equal(theta1, theta, 12)
        npt.assert_array_less(parallax1, parallax)
        npt.assert_almost_equal(muphistar1, muphistar, 12)
        npt.assert_almost_equal(mutheta1, mutheta, 12)
        npt.assert_array_less(pmr1, pmr)

        # For propagation into the past check that In the absence of proper motion and
        # for positive radial velocity and positive parallax at the current epoch: the
        # parallax always increases, the radial proper motion always increases.
        phi2, theta2, parallax2, muphistar2, mutheta2, pmr2 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t2
        )
        npt.assert_almost_equal(phi2, phi, 12)
        npt.assert_almost_equal(theta2, theta, 12)
        npt.assert_array_less(parallax, parallax2)
        npt.assert_almost_equal(muphistar2, muphistar, 12)
        npt.assert_almost_equal(mutheta2, mutheta, 12)
        npt.assert_array_less(pmr, pmr2)

        # For propagation into the future check that In the absence of proper motion and
        # for negative radial velocity and positive parallax at the current epoch: the
        # parallax always increases, the radial proper motion always decreases.
        vrad = -1.0 - 70 * rng.uniform(size=nTests)
        pmr = vrad * parallax / au_km_year_per_sec
        phi1, theta1, parallax1, muphistar1, mutheta1, pmr1 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t1
        )
        npt.assert_almost_equal(phi1, phi, 12)
        npt.assert_almost_equal(theta1, theta, 12)
        npt.assert_array_less(parallax, parallax1)
        npt.assert_almost_equal(muphistar1, muphistar, 12)
        npt.assert_almost_equal(mutheta1, mutheta, 12)
        npt.assert_array_less(pmr1, pmr)

        # For propagation into the past check that In the absence of proper motion and
        # for negative radial velocity and positive parallax at the current epoch: the
        # parallax always decreases, and the radial proper motion always increases.
        phi2, theta2, parallax2, muphistar2, mutheta2, pmr2 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t2
        )
        npt.assert_almost_equal(phi2, phi, 12)
        npt.assert_almost_equal(theta2, theta, 12)
        npt.assert_array_less(parallax2, parallax)
        npt.assert_almost_equal(muphistar2, muphistar, 12)
        npt.assert_almost_equal(mutheta2, mutheta, 12)
        npt.assert_array_less(pmr, pmr2)

        # Check that angular distance between old and new position is close to that
        # obtained from a naive epoch propagation.
        nTests = 10000
        phi = 2.0 * np.pi * rng.uniform(size=nTests)
        theta = np.arcsin(-1 + 2 * rng.uniform(size=nTests))
        parallax = rng.uniform(size=nTests) * 20.0
        muphistar = rng.uniform(size=nTests) * 200 - 100
        mutheta = rng.uniform(size=nTests) * 200 - 100
        vrad = rng.uniform(size=nTests) * 400 - 200
        phi1, theta1, parallax1, muphistar1, mutheta1, pmr1 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t1
        )
        phi2, theta2, parallax2, muphistar2, mutheta2, pmr2 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t2
        )
        rho1 = angular_distance(phi, theta, phi1, theta1) * radtomas
        rho2 = angular_distance(phi, theta, phi2, theta2) * radtomas
        rho_naive1 = np.sqrt(muphistar**2 + mutheta**2) * abs(t1 - t0)
        rho_naive2 = np.sqrt(muphistar**2 + mutheta**2) * abs(t2 - t0)
        npt.assert_allclose(rho_naive1, rho1, rtol=0.1)
        npt.assert_allclose(rho_naive2, rho2, rtol=0.1)

        # Check that for zero parallax the propgation still works, leaving parallax
        # unchanged.
        nTests = 100
        phi = 2.0 * np.pi * rng.uniform(size=nTests)
        theta = np.arcsin(-1 + 2 * rng.uniform(size=nTests))
        parallax = np.zeros_like(phi)
        muphistar = rng.uniform(size=nTests) * 2 - 1
        mutheta = rng.uniform(size=nTests) * 2 - 1
        vrad = rng.uniform(size=nTests) * 40 - 20
        phi1, theta1, parallax1, muphistar1, mutheta1, pmr1 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t1
        )
        phi2, theta2, parallax2, muphistar2, mutheta2, pmr2 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t2
        )
        npt.assert_equal(parallax1, 0.0)
        npt.assert_equal(parallax2, 0.0)
        npt.assert_array_less(0.0, pmr1)
        npt.assert_array_less(pmr2, 0.0)

        # For propagation into the future check that In the absence of proper motion and
        # for positive radial velocity and negative parallax at the current epoch: the
        # parallax always decreases, the radial proper motion decreases.
        nTests = 100
        phi = 2.0 * np.pi * rng.uniform(size=nTests)
        theta = np.arcsin(-1 + 2 * rng.uniform(size=nTests))
        parallax = -rng.uniform(size=nTests) * 20.0 - 1.0
        muphistar = np.zeros_like(phi)
        mutheta = np.zeros_like(phi)
        vrad = rng.uniform(size=nTests) * 50.0 + 1.0
        pmr = vrad * parallax / au_km_year_per_sec
        phi1, theta1, parallax1, muphistar1, mutheta1, pmr1 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t1
        )
        npt.assert_almost_equal(phi1, phi, 12)
        npt.assert_almost_equal(theta1, theta, 12)
        npt.assert_array_less(parallax1, parallax)
        npt.assert_almost_equal(muphistar1, muphistar, 12)
        npt.assert_almost_equal(mutheta1, mutheta, 12)
        npt.assert_array_less(pmr1, pmr)

        # For propagation into the past check that In the absence of proper motion and
        # for positive radial velocity and negative parallax at the current epoch: the
        # parallax always increases, the radial proper motion always increases.
        phi2, theta2, parallax2, muphistar2, mutheta2, pmr2 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t2
        )
        npt.assert_almost_equal(phi2, phi, 12)
        npt.assert_almost_equal(theta2, theta, 12)
        npt.assert_array_less(parallax, parallax2)
        npt.assert_almost_equal(muphistar2, muphistar, 12)
        npt.assert_almost_equal(mutheta2, mutheta, 12)
        npt.assert_array_less(pmr, pmr2)

        # For propagation into the future check that In the absence of proper motion and
        # for negative radial velocity and negative parallax at the current epoch: the
        # parallax always increases, the radial proper motion always decreases.
        vrad = -1.0 - 70 * rng.uniform(size=nTests)
        pmr = vrad * parallax / au_km_year_per_sec
        phi1, theta1, parallax1, muphistar1, mutheta1, pmr1 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t1
        )
        npt.assert_almost_equal(phi1, phi, 12)
        npt.assert_almost_equal(theta1, theta, 12)
        npt.assert_array_less(parallax, parallax1)
        npt.assert_almost_equal(muphistar1, muphistar, 12)
        npt.assert_almost_equal(mutheta1, mutheta, 12)
        npt.assert_array_less(pmr1, pmr)

        # For propagation into the past check that In the absence of proper motion and
        # for negative radial velocity and negative parallax at the current epoch: the
        # parallax always decreases, and the radial proper motion always increases.
        phi2, theta2, parallax2, muphistar2, mutheta2, pmr2 = ep.propagate_astrometry(
            phi, theta, parallax, muphistar, mutheta, vrad, t0, t2
        )
        npt.assert_almost_equal(phi2, phi, 12)
        npt.assert_almost_equal(theta2, theta, 12)
        npt.assert_array_less(parallax2, parallax)
        npt.assert_almost_equal(muphistar2, muphistar, 12)
        npt.assert_almost_equal(mutheta2, mutheta, 12)
        npt.assert_array_less(pmr, pmr2)

    def test_propagate_astrometry_and_covariance_matrix(self):
        """
        Verify correct working of propagate_astrometry_and_covariance_matrix().
        """

        ep = EpochPropagation()
        t0 = 2015.5
        t1 = 2020.5
        t2 = 2010.0

        nTests = 10000
        a0 = np.zeros((6, nTests))
        a0[0] = 2.0 * np.pi * rng.uniform(size=nTests)
        a0[1] = np.arcsin(-1 + 2 * rng.uniform(size=nTests))
        a0[2] = rng.uniform(size=nTests) * 20.0
        a0[3] = rng.uniform(size=nTests) * 200 - 100
        a0[4] = rng.uniform(size=nTests) * 200 - 100
        a0[5] = rng.uniform(size=nTests) * 400 - 200

        covmat = np.array(
            [1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        )
        vrad_error = 2.0
        c0 = construct_covariance_matrix(
            np.tile(covmat, (nTests, covmat.size)),
            a0[2],
            a0[5],
            np.tile(vrad_error, nTests),
        )

        for i in range(nTests):
            self.assertTrue(self._is_pos_def(c0[i]))

        a1, c1 = ep.propagate_astrometry_and_covariance_matrix(a0, c0, t0, t1)
        a2, c2 = ep.propagate_astrometry_and_covariance_matrix(a0, c0, t0, t2)
        phi1, theta1, parallax1, muphistar1, mutheta1, pmr1 = ep.propagate_astrometry(
            a0[0], a0[1], a0[2], a0[3], a0[4], a0[5], t0, t1
        )
        phi2, theta2, parallax2, muphistar2, mutheta2, pmr2 = ep.propagate_astrometry(
            a0[0], a0[1], a0[2], a0[3], a0[4], a0[5], t0, t2
        )

        npt.assert_almost_equal(phi1, a1[0], 12)
        npt.assert_almost_equal(theta1, a1[1], 12)
        npt.assert_almost_equal(parallax1, a1[2], 12)
        npt.assert_almost_equal(muphistar1, a1[3], 12)
        npt.assert_almost_equal(mutheta1, a1[4], 12)
        npt.assert_almost_equal(pmr1, a1[5], 12)

        npt.assert_almost_equal(phi2, a2[0], 12)
        npt.assert_almost_equal(theta2, a2[1], 12)
        npt.assert_almost_equal(parallax2, a2[2], 12)
        npt.assert_almost_equal(muphistar2, a2[3], 12)
        npt.assert_almost_equal(mutheta2, a2[4], 12)
        npt.assert_almost_equal(pmr2, a2[5], 12)

        for i in range(nTests):
            self.assertTrue(self._is_pos_def(c1[i]))
            self.assertTrue(self._is_pos_def(c2[i]))

        a1, c1 = ep.propagate_astrometry_and_covariance_matrix(a0[:, 0], c0[0], t0, t1)
        a2, c2 = ep.propagate_astrometry_and_covariance_matrix(a0[:, 0], c0[0], t0, t2)
        phi1, theta1, parallax1, muphistar1, mutheta1, pmr1 = ep.propagate_astrometry(
            a0[0, 0], a0[1, 0], a0[2, 0], a0[3, 0], a0[4, 0], a0[5, 0], t0, t1
        )
        phi2, theta2, parallax2, muphistar2, mutheta2, pmr2 = ep.propagate_astrometry(
            a0[0, 0], a0[1, 0], a0[2, 0], a0[3, 0], a0[4, 0], a0[5, 0], t0, t2
        )

        npt.assert_almost_equal(phi1, a1[0], 12)
        npt.assert_almost_equal(theta1, a1[1], 12)
        npt.assert_almost_equal(parallax1, a1[2], 12)
        npt.assert_almost_equal(muphistar1, a1[3], 12)
        npt.assert_almost_equal(mutheta1, a1[4], 12)
        npt.assert_almost_equal(pmr1, a1[5], 12)

        npt.assert_almost_equal(phi2, a2[0], 12)
        npt.assert_almost_equal(theta2, a2[1], 12)
        npt.assert_almost_equal(parallax2, a2[2], 12)
        npt.assert_almost_equal(muphistar2, a2[3], 12)
        npt.assert_almost_equal(mutheta2, a2[4], 12)
        npt.assert_almost_equal(pmr2, a2[5], 12)

        self.assertTrue(self._is_pos_def(c1))
        self.assertTrue(self._is_pos_def(c2))

    def test_angular_distance(self):
        """
        Verify correct working of angular_distance function.

        Returns
        -------
        Nothing
        """
        ra1 = np.deg2rad([0, 0, 0, 35, 35, 65, 200, 80])
        dec1 = np.deg2rad([0, 0, 0, 0, 0, 30, 90, 0])
        ra2 = np.deg2rad([0, 0, 90, 35, 215, 245, 200, 350])
        dec2 = np.deg2rad([0, 90, 0, 10, 0, -30, 80, 0])

        rho, theta = angular_distance(ra1, dec1, ra2, dec2, return_posangle=True)
        expected_rho = np.deg2rad([0, 90, 90, 10, 180, 180, 10, 90])
        expected_theta = np.deg2rad([0, 0, 90, 0, 0, 0, 180, 270])

        npt.assert_array_almost_equal(rho, expected_rho, 12)
        npt.assert_array_almost_equal(theta, expected_theta, 12)

    @staticmethod
    def _generateRandomCovarianceMatrices(num):
        """
        Generate random covariance matrices through the transformation of diagonal
        positive definite matrices.

        Parameters
        ----------
        num : int
            Number of matrices to generate.

        Returns
        -------
        sigma1 : float
            Square root of variance of first variable.
        sigma2 : float
            Square root of variance of second variable.
        rho12 : float
            Correlation coefficient between errors on variables 1 and 2
        """
        varDiag1 = rng.uniform(size=num) * 99.0 + 1.0
        varDiag2 = rng.uniform(size=num) * 99.0 + 1.0
        rotAngle = 2.0 * np.pi * rng.uniform(size=num)
        sigma1 = np.zeros(num)
        sigma2 = np.zeros(num)
        rho12 = np.zeros(num)
        for i in range(num):
            rotationMatrix = np.array(
                [
                    [np.cos(rotAngle[i]), np.sin(rotAngle[i])],
                    [-np.sin(rotAngle[i]), np.cos(rotAngle[i])],
                ]
            )
            covMatrix = np.dot(
                rotationMatrix,
                np.dot(
                    np.diag([varDiag1[i], varDiag2[i]]), np.transpose(rotationMatrix)
                ),
            )
            sigma1[i] = np.sqrt(covMatrix[0, 0])
            sigma2[i] = np.sqrt(covMatrix[1, 1])
            rho12[i] = covMatrix[0, 1] / (sigma1[i] * sigma2[i])

        return sigma1, sigma2, rho12

    @staticmethod
    def _is_pos_def(A):
        """
        Check if matrix A is positive definite. Code from
        https://stackoverflow.com/questions/16266720/find-out-if-matrix-is-positive-definite-with-numpy
        """
        if np.allclose(A, A.T, rtol=1e-10, atol=1e-12):
            try:
                np.linalg.cholesky(A)
                return True
            except np.linalg.LinAlgError:
                return False
        else:
            return False



"""
Provides transformations between Gaia and a few other photometric systems.

The transformations are taken from the `Gaia DR3 documentation. <https://gea.esac.esa.int/archive/documentation/GDR3/Data_processing/chap_cu5pho/cu5pho_sec_photSystem/cu5pho_ssec_photRelations.html>`_
"""
import numpy as np

__all__ = [
    "gbrminv_from_vminic",
    "gminic_from_vminic",
    "grvs_from_gmingrp",
    "grvs_from_vminic",
]


def gbrminv_from_vminic(vminic):
    r"""
    Transformation from Johnson-Cousins :math:`V-I_\mathrm{c}` to Gaia magnitudes.

    Calculate :math:`G-V`, :math:`G_\mathrm{BP}-V`, :math:`G_\mathrm{RP}-V`, and
    :math:`G_\mathrm{BP}-G_\mathrm{RP}` from the input values of
    :math:`V-I_\mathrm{c}`

    Parameters
    ----------
    vminic : ndarray, float
        Value(s) of :math:`V-I_\mathrm{c}` for which to calculate the Gaia magnitudes

    Returns
    -------
    gminv, gbpminv, grpmin, gbpmingrp : ndarray, float
        The value(s) of :math:`G-V`, :math:`G_\mathrm{BP}-V`,
        :math:`G_\mathrm{RP}-V`, and :math:`G_\mathrm{BP}-G_\mathrm{RP}`
    """
    vminic2 = vminic * vminic
    vminic3 = vminic2 * vminic
    vminic4 = vminic3 * vminic
    gminv = (
        -0.01597
        - 0.02809 * vminic
        - 0.2483 * vminic2
        + 0.03656 * vminic3
        - 0.002939 * vminic4
    )
    gbpminv = -0.0143 + 0.3564 * vminic - 0.1332 * vminic2 + 0.01212 * vminic3
    grpminv = 0.01868 - 0.9028 * vminic - 0.005321 * vminic2 - 0.004186 * vminic3
    gbpmingrp = -0.03298 + 1.259 * vminic - 0.1279 * vminic2 + 0.01631 * vminic3
    return gminv, gbpminv, grpminv, gbpmingrp


def gminic_from_vminic(vminic):
    r"""
    Transformation from Johnson-Cousins :math:`V-I_\mathrm{c}` to :math:`G-I_\mathrm{c}`.

    Calculate :math:`G-I_\mathrm{c}` from the input values of :math:`V-I_\mathrm{c}`

    Parameters
    ----------
    vminic : ndarray, float
        Value(s) of :math:`V-I_\mathrm{c}` for which to calculate the :math:`G-I_\mathrm{c}` values.

    Returns
    -------
    gminic : ndarray, float
        The value(s) of :math:`G-I_\mathrm{c}`.
    """
    return 0.01753 + 0.76 * vminic - 0.0991 * vminic**2


def grvs_from_gmingrp(gminrp):
    r"""
    Calculate :math:`G_\mathrm{RVS}` from the input value(s) of :math:`G-G_\mathrm{RP}`.
    Use the formulae presented in `Sartoretti et al. (2022)
    <https://ui.adsabs.harvard.edu/abs/2022arXiv220605725S/abstract>`_.

    Parameters
    ----------
    gminrp : ndarray, float
        Value(s) of :math:`G-G_\mathrm{RP}` from which to calculate
        :math:`G_\mathrm{RVS}`. Must be in the range :math:`-0.15\leq
        G-G_\mathrm{RVS}\leq 1.7`

    Returns
    -------
    gminrvs : ndarray, float
        The values of :math:`G_\mathrm{RVS}`. NANs are returned for
        :math:`G-G_\mathrm{RP}` values outside the above quoted range.
    """
    gmrp = np.array(gminrp)
    gminrvs = np.zeros_like(gmrp)
    blue = (gmrp >= -0.15) & (gmrp <= 1.2)
    red = (gmrp > 1.2) & (gmrp <= 1.7)
    gminrvs[blue] = (
        -0.0397
        - 0.2852 * gmrp[blue]
        - 0.0330 * np.power(gmrp[blue], 2)
        - 0.0867 * np.power(gmrp[blue], 3)
    )
    gminrvs[red] = (
        -4.0618
        + 10.0187 * gmrp[red]
        - 9.0532 * np.power(gmrp[red], 2)
        + 2.6089 * np.power(gmrp[red], 3)
    )
    gminrvs[np.logical_not(blue | red)] = np.nan
    return gminrvs


def grvs_from_vminic(vminic):
    r"""
    Calculate :math:`G_\mathrm{RVS}` from the input value(s) of :math:`V-I_\mathrm{c}`.

    Parameters
    ----------
    vminic : ndarray, float
        Value(s) of :math:`V-I_\mathrm{c}` from which to calculate
        :math:`G_\mathrm{RVS}`.

    Returns
    -------
    gminrvs : ndarray, float
        The values of :math:`G_\mathrm{RVS}`. NANs are returned for
        :math:`V-I_\mathrm{c}` values corresponding to :math:`G-G_\mathrm{RP}` values
        outside the valid range listed in `Sartoretti et al. (2022)
        <https://ui.adsabs.harvard.edu/abs/2022arXiv220605725S/abstract>`_.
    """
    gminv, _, grpminv, _ = gbrminv_from_vminic(vminic)
    return grvs_from_gmingrp(gminv - grpminv)



from setuptools import find_packages, setup

setup(
    name="PyGaia",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={
        "": ["LICENSE", "AUTHORS.md", "HISTORY.md", "INSTALL.md"],
        "pygaia.errors": ["data/*"],
    },
    include_package_data=True,
)



"""
Gaia simulation tools.

This package provides classes and functions for simulating the uncertainties
expected on the Gaia catalogue data. The formulae and tables are taken from the
`Gaia science performance pages
<https://www.cosmos.esa.int/web/gaia/science-performance>`_.
"""

from . import astrometric
from . import photometric
from . import spectroscopic
from . import utils



r"""
Provides utility functions for photometric transformations. Values of :math:`M_V` and :math:`V-I_\mathrm{c}` are from `Pickles (1998) <https://ui.adsabs.harvard.edu/abs/1998PASP..110..863P/abstract>`_.
"""
from pygaia.photometry.transformations import gbrminv_from_vminic

__all__ = ["vminic_from_spt", "vabs_from_spt", "gabs_from_spt"]

_spt_to_vminic_vabs = {
    "B0V": (-0.31, -3.5),
    "B1V": (-0.24, -2.7),
    "B5V": (-0.08, 0.0),
    "A0V": (0.01, 0.0),
    "A5V": (0.16, 1.69),
    "F0V": (0.38, 2.98),
    "G0V": (0.67, 4.24),
    "G2V": (0.72, 4.7),
    "G5V": (0.74, 4.78),
    "K0V": (0.87, 5.58),
    "K1IIIMP": (0.99, 1.53),
    "K4V": (1.23, 7.21),
    "K1III": (1.04, 2.16),
    "M0V": (1.71, 8.62),
    "M2V": (2.02, 9.48),
    "M6V": (3.69, 14.2),
    "M0III": (1.65, -0.66),
    "B0I": (-0.22, -7.1),
    "B1I": (-0.16, -6.7),
}


def vminic_from_spt(spt):
    r"""
    Obtain :math:`(V-I_\mathrm{c})` for the input spectral type.

    Parameters
    ----------
    spt : str
        String representing the spectral type of the star.

    Returns
    -------
    vmini : float
        The value of :math:`(V-I_\mathrm{c})`.

    Raises
    ------
    ValueError
        When an unknown spectral type is specified.
    """
    if spt in _spt_to_vminic_vabs:
        return _spt_to_vminic_vabs[spt][0]
    else:
        message = "Unknown spectral type. Allowed values are: "
        for key in _spt_to_vminic_vabs.keys():
            message += key + " "
        raise ValueError(message)


def vabs_from_spt(spt):
    r"""
    Obtain :math:`M_V` (absolute magnitude in :math:`V`-band) for the input spectral type.

    Parameters
    ----------
    spt : str
        String representing the spectral type of the star.

    Returns
    -------
    vabs : float
        The value of :math:`M_V`.

    Raises
    ------
    ValueError
        When an unknown spectral type is specified.
    """
    if spt in _spt_to_vminic_vabs:
        return _spt_to_vminic_vabs[spt][1]
    else:
        message = "Unknown spectral type. Allowed values are: "
        for key in _spt_to_vminic_vabs.keys():
            message += key + " "
        raise ValueError(message)


def gabs_from_spt(spt):
    r"""
    Obtain :math:`M_G` (absolute magnitude in :math:`G`-band) for the input spectral type.

    Parameters
    ----------
    spt : str
        String representing the spectral type of the star.

    Returns
    -------
    gabs : float
        The value of :math:`M_G`.

    Raises
    ------
    ValueError
        When an unknown spectral type is specified.
    """
    if spt in _spt_to_vminic_vabs:
        vminic = vminic_from_spt(spt)
        gminv, _, _, _ = gbrminv_from_vminic(vminic)
        return vabs_from_spt(spt) + gminv
    else:
        message = "Unknown spectral type. Allowed values are: "
        for key in _spt_to_vminic_vabs.keys():
            message += key + " "
        raise ValueError(message)
